"""
_CountUndeletedBlocksByWorkflow_

MySQL implementation of Workflows.CountUndeletedBlocksByWorkflow

Retrieves a list of workflows and the relative undeleted blocks counters,
"""


from WMCore.Database.DBFormatter import DBFormatter


class CountUndeletedBlocksByWorkflow(DBFormatter):
    """
    Retrieves a list of all workflows and the relative undeleted blocks counters

    The structure returned:
    [{'count': 6,
      'deleted': 0,
      'name': 'PromptReco_Run351572_HcalNZS_Tier0_REPLAY_2022_ID220531142559_v425_220531_1430'},
     {'count': 8,
      'deleted': 0,
      'name': 'PromptReco_Run351572_NoBPTX_Tier0_REPLAY_2022_ID220531142559_v425_220531_1430'},
      ...]
    """
    sql = """
    SELECT
        dbsbuffer_workflow.name,
        COUNT(DISTINCT dbsbuffer_block.blockname) as count
    FROM dbsbuffer_block
    INNER JOIN dbsbuffer_file ON
        dbsbuffer_file.block_id = dbsbuffer_block.id
    INNER JOIN dbsbuffer_workflow ON
        dbsbuffer_workflow.id = dbsbuffer_file.workflow
    WHERE dbsbuffer_block.deleted=0
    GROUP BY
        dbsbuffer_workflow.name
    """

    def execute(self, conn=None, transaction=False, returnCursor=False):
        """
        Executing the current sql query.
        :param conn:        A current database connection to be used if existing
        :param transaction: A current database transaction to be used if existing
        :return:            A list of dictionaries one record for each database line returned
        """
        dictResults = DBFormatter.formatDict(self, self.dbi.processData(self.sql, conn=conn,
                                                                        transaction=transaction))
        return dictResults



#!/usr/bin/env python
"""
_Parentage_

MySQL implementation of Fileset.Parentage

"""
__all__ = []



from WMCore.Database.DBFormatter import DBFormatter

class Parentage(DBFormatter):
    sql = """insert into wmbs_fileset_parent (child, parent)
                values ((select id from wmbs_fileset where name = :child),
                (select id from wmbs_fileset where name = :parent))"""

    def getBinds(self, child = 0, parent = 0):
        return self.dbi.buildbinds(child, 'child', self.dbi.buildbinds(parent, 'parent'))

    def execute(self, child = 0, parent = 0, conn = None, transaction = False):
        result = self.dbi.processData(self.sql, self.getBinds(child, parent),
                         conn = conn, transaction = transaction)
        return self.format(result)



"""
_GetUserId_

MySQL implementation of Users.GetUserId
"""

__all__ = []



import time
from WMCore.Database.DBFormatter import DBFormatter

class GetUserId(DBFormatter):
    sql = """SELECT id FROM wmbs_users
             WHERE cert_dn = :dn
             AND group_name = :gr
             AND role_name = :role
          """

    def execute(self, dn = None, group_name = '', role_name = '', conn = None, transaction = False):
        binds = {"dn": dn, "gr": group_name, "role": role_name}
        result = self.dbi.processData(self.sql, binds, conn = conn,
                                      transaction = transaction)
        id = self.format(result)
        if len(id) > 0:
            return int(id[0][0])

        return None



"""
_GetDeletableBlocks_

MySQL implementation of PhEDExInjector.GetDeletableBlocks

Retrieve a list of blocks that can be deleted,
including their location and the sites they
are subscribed too

"""

from __future__ import division
from __future__ import print_function

from WMCore.Database.DBFormatter import DBFormatter


class GetDeletableBlocks(DBFormatter):
    # Retrieve a list of blocks that can be deleted:
    #   - workflow for all files in block completed
    #   - subscription made for dataset is copy+delete
    #   - subscription has been made in PhEDEx
    #   - blocks hasn't been deleted yet

    # FIXME: returns SE, should switch to PNN soon

    sql = """SELECT dbsbuffer_block.blockname,
                    dbsbuffer_location.pnn,
                    dbsbuffer_dataset.path,
                    dbsbuffer_dataset_subscription.site
             FROM dbsbuffer_dataset_subscription
             INNER JOIN dbsbuffer_dataset ON
               dbsbuffer_dataset.id = dbsbuffer_dataset_subscription.dataset_id
             INNER JOIN dbsbuffer_block ON
               dbsbuffer_block.dataset_id = dbsbuffer_dataset_subscription.dataset_id
             INNER JOIN dbsbuffer_file ON
               dbsbuffer_file.block_id = dbsbuffer_block.id
             INNER JOIN dbsbuffer_workflow ON
               dbsbuffer_workflow.id = dbsbuffer_file.workflow
             INNER JOIN dbsbuffer_location ON
               dbsbuffer_location.id = dbsbuffer_block.location
             LEFT OUTER JOIN wmbs_workflow ON
               wmbs_workflow.name = dbsbuffer_workflow.name
             WHERE dbsbuffer_dataset_subscription.delete_blocks = 1
             AND dbsbuffer_dataset_subscription.subscribed = 1
             AND dbsbuffer_block.status = 'Closed'
             AND dbsbuffer_block.deleted = 0
             GROUP BY dbsbuffer_block.blockname,
                      dbsbuffer_location.pnn,
                      dbsbuffer_dataset.path,
                      dbsbuffer_dataset_subscription.site
             HAVING COUNT(*) = SUM(dbsbuffer_workflow.completed)
             AND COUNT(wmbs_workflow.name) = 0
             """

    def execute(self, conn=None, transaction=False):

        results = self.dbi.processData(self.sql, conn=conn,
                                       transaction=transaction)[0].fetchall()

        blockDict = {}
        for result in results:

            blockName = result[0]
            location = result[1]
            dataset = result[2]
            site = result[3]

            if blockName not in blockDict:
                blockDict[blockName] = {'location': location,
                                        'dataset': dataset,
                                        'sites': set()}

            blockDict[blockName]['sites'].add(site)

        return blockDict



#!/usr/bin/env python
"""
_Exists_

Oracle implementation of JobGroup.Exists
"""

__all__ = []



from WMCore.WMBS.MySQL.JobGroup.Exists import Exists as ExistsJobGroupMySQL

class Exists(ExistsJobGroupMySQL):
    pass



#!/usr/bin/env python
"""
_SetFWJRPath_

Oracle implementation of Jobs.SetFWJRPath
"""




from WMCore.WMBS.MySQL.Jobs.SetFWJRPath import SetFWJRPath as MySQLSetFWJRPath

class SetFWJRPath(MySQLSetFWJRPath):
    """
    Identical to MySQL version.
    """
    pass



#!/usr/bin/env python
"""
_RetriedJobs_

Oracle implementation of Workflow.FailedJobs
"""




from WMCore.WMBS.MySQL.Workflow.RetriedJobs import RetriedJobs as MySQLRetriedJobs

class RetriedJobs(MySQLRetriedJobs):
    pass



#!/usr/bin/env python
"""
_SiblingSubscriptionsComplete_

Oracle implementation of Subscription.SiblingSubscriptionsComplete
"""

from WMCore.WMBS.MySQL.Subscriptions.SiblingSubscriptionsComplete import \
    SiblingSubscriptionsComplete as SiblingCompleteMySQL

class SiblingSubscriptionsComplete(SiblingCompleteMySQL):
    pass



#!/usr/bin/env python
"""
_Jobs_

Oracle implementation of Subscriptions.Jobs

Return a list of all jobs that exist for a subscription.
"""

__all__ = []



from WMCore.WMBS.MySQL.Subscriptions.Jobs import Jobs as JobsMySQL

class Jobs(JobsMySQL):
    sql = JobsMySQL.sql



#!/usr/bin/env python
"""
_Exists_

MySQL implementation of Jobs.Exists
"""

__all__ = []



from WMCore.Database.DBFormatter import DBFormatter

class Exists(DBFormatter):
    sql = "select id from wmbs_job where name = :name"

    def format(self, result):
        result = DBFormatter.format(self, result)

        if len(result) == 0:
            return False
        else:
            return result[0][0]

    def getBinds(self, name):
        return self.dbi.buildbinds(self.dbi.makelist(name), "name")

    def execute(self, name, conn = None, transaction = False):
        result = self.dbi.processData(self.sql, self.getBinds(name),
                                      conn = conn, transaction = transaction)
        return self.format(result)



#!/usr/bin/env python
"""
_ScriptInterface_


API definition for RuntimeScripts that will be called in the job area.


"""
from builtins import object


class ScriptInterface(object):
    """
    _ScriptInterface_


    """
    def __init__(self):
        self.stepSpace = None

    def __call__(self):
        """
        _operator_

        Override this to perform the required tasks and return an appropriate
        exit code

        """
        return 0



"""
_GetCompletedBlocks_

Oracle implementation of RucioInjector.GetComletedBlocks

"""


from WMComponent.RucioInjector.Database.MySQL.GetCompletedBlocks import GetCompletedBlocks as MySQLGetCompletedBlocks


class GetCompletedBlocks(MySQLGetCompletedBlocks):
    pass



#!/usr/bin/env python
"""
_PromptReco_

Standard PromptReco workflow.
"""
from __future__ import division
from future.utils import viewitems

from Utils.Utilities import makeList, strToBool
from WMCore.Lexicon import procstringT0
from WMCore.WMSpec.StdSpecs.DataProcessing import DataProcessing


class PromptRecoWorkloadFactory(DataProcessing):
    """
    _PromptRecoWorkloadFactory_

    Stamp out PromptReco workflows.
    """

    def __init__(self):
        super(PromptRecoWorkloadFactory, self).__init__()
        self.procJobSplitArgs = {}
        self.skimJobSplitArgs = {}
        self.inputProcessedDataset = None
        self.inputPrimaryDataset = None
        self.eventsPerJob = None
        self.inputDataTier = None

    def buildWorkload(self):
        """
        _buildWorkload_

        Build the workload given all of the input parameters.

        Not that there will be LogCollect tasks created for each processing
        task and Cleanup tasks created for each merge task.

        """
        (self.inputPrimaryDataset, self.inputProcessedDataset,
         self.inputDataTier) = self.inputDataset[1:].split("/")

        workload = self.createWorkload()
        workload.setDashboardActivity("t0")
        workload.setWorkQueueSplitPolicy("Block", self.procJobSplitAlgo,
                                         self.procJobSplitArgs)

        cmsswStepType = "CMSSW"
        taskType = "Processing"

        recoOutputs = []
        for dataTier in self.writeTiers:
            recoOutputs.append({'dataTier': dataTier,
                                'eventContent': dataTier,
                                'moduleLabel': "write_%s" % dataTier})

        recoTask = workload.newTask("Reco")

        scenarioArgs = {'globalTag': self.globalTag,
                        'skims': self.alcaSkims,
                        'PhysicsSkims': self.physicsSkims,
                        'dqmSeq': self.dqmSequences,
                        'outputs': recoOutputs}
        if self.globalTagConnect:
            scenarioArgs['globalTagConnect'] = self.globalTagConnect

        recoOutMods = self.setupProcessingTask(recoTask, taskType, self.inputDataset,
                                               scenarioName=self.procScenario,
                                               scenarioFunc="promptReco",
                                               scenarioArgs=scenarioArgs,
                                               splitAlgo=self.procJobSplitAlgo,
                                               splitArgs=self.procJobSplitArgs,
                                               stepType=cmsswStepType,
                                               forceUnmerged=[
                                                   "write_ALCARECO"] if 'ALCARECO' in self.writeTiers else False)
        if self.doLogCollect:
            self.addLogCollectTask(recoTask)

        recoMergeTasks = {}
        for recoOutLabel, recoOutInfo in viewitems(recoOutMods):
            if recoOutInfo['dataTier'] != "ALCARECO":
                mergeTask = self.addMergeTask(recoTask, self.procJobSplitAlgo, recoOutLabel,
                                              doLogCollect=self.doLogCollect)
                recoMergeTasks[recoOutInfo['dataTier']] = mergeTask

            else:
                alcaTask = recoTask.addTask("AlcaSkim")
                alcaTaskConf = {'Multicore': 1, 'EventStreams': 0}
                scenarioArgs = {'globalTag': self.globalTag,
                                'skims': self.alcaSkims,
                                'primaryDataset': self.inputPrimaryDataset}
                if self.globalTagConnect:
                    scenarioArgs['globalTagConnect'] = self.globalTagConnect

                alcaOutMods = self.setupProcessingTask(alcaTask, taskType,
                                                       inputStep=recoTask.getStep("cmsRun1"),
                                                       inputModule=recoOutLabel,
                                                       scenarioName=self.procScenario,
                                                       scenarioFunc="alcaSkim",
                                                       scenarioArgs=scenarioArgs,
                                                       splitAlgo="ParentlessMergeBySize",
                                                       splitArgs={"max_merge_size": self.maxMergeSize,
                                                                  "min_merge_size": self.minMergeSize,
                                                                  "max_merge_events": self.maxMergeEvents},
                                                       stepType=cmsswStepType,
                                                       taskConf=alcaTaskConf)
                if self.doLogCollect:
                    self.addLogCollectTask(alcaTask, taskName="AlcaSkimLogCollect")
                self.addCleanupTask(recoTask, recoOutLabel, dataTier=recoOutInfo['dataTier'])

                for alcaOutLabel in alcaOutMods:
                    self.addMergeTask(alcaTask, self.procJobSplitAlgo, alcaOutLabel,
                                      doLogCollect=self.doLogCollect)

        # setting the parameters which need to be set for all the tasks
        # sets acquisitionEra, processingVersion, processingString
        workload.setTaskPropertiesFromWorkload()

        # set the LFN bases (normally done by request manager)
        # also pass runNumber (workload evaluates it)
        workload.setLFNBase(self.mergedLFNBase, self.unmergedLFNBase,
                            runNumber=self.runNumber)

        return workload

    def __call__(self, workloadName, arguments):
        """
        _call_

        Create a ReReco workload with the given parameters.
        """
        DataProcessing.__call__(self, workloadName, arguments)

        # These are mostly place holders because the job splitting algo and
        # parameters will be updated after the workflow has been created.
        self.procJobSplitArgs = {}
        if self.procJobSplitAlgo in ["EventBased", "EventAwareLumiBased"]:
            if self.eventsPerJob is None:
                self.eventsPerJob = int((8.0 * 3600.0) / self.timePerEvent)
            self.procJobSplitArgs["events_per_job"] = self.eventsPerJob
            if self.procJobSplitAlgo == "EventAwareLumiBased":
                self.procJobSplitArgs["job_time_limit"] = 96 * 3600  # 4 days in seconds
                self.procJobSplitArgs["allowCreationFailure"] = False
        elif self.procJobSplitAlgo == "LumiBased":
            self.procJobSplitArgs["lumis_per_job"] = self.lumisPerJob
        elif self.procJobSplitAlgo == "FileBased":
            self.procJobSplitArgs["files_per_job"] = self.filesPerJob

        self.skimJobSplitArgs = {}
        if self.skimJobSplitAlgo in ["EventBased", "EventAwareLumiBased"]:
            if self.eventsPerJob is None:
                self.eventsPerJob = int((8.0 * 3600.0) / self.timePerEvent)
            if self.skimJobSplitAlgo == "EventAwareLumiBased":
                self.skimJobSplitArgs["job_time_limit"] = 48 * 3600  # 2 days
                self.skimJobSplitArgs["allowCreationFailure"] = False
            self.skimJobSplitArgs["events_per_job"] = self.eventsPerJob
        elif self.skimJobSplitAlgo == "LumiBased":
            self.skimJobSplitArgs["lumis_per_job"] = self.lumisPerJob
        elif self.skimJobSplitAlgo == "FileBased":
            self.skimJobSplitArgs["files_per_job"] = self.filesPerJob
        self.skimJobSplitArgs = arguments.get("SkimJobSplitArgs",
                                              {"files_per_job": 1,
                                               "include_parents": True})

        return self.buildWorkload()

    @staticmethod
    def getWorkloadCreateArgs():
        baseArgs = DataProcessing.getWorkloadCreateArgs()
        specArgs = {"RequestType": {"default": "PromptReco", "optional": True},
                    "ConfigCacheID": {"optional": True, "null": True},
                    "Scenario": {"default": None, "optional": False,
                                 "attr": "procScenario", "null": False},
                    "ProcessingString": {"default": "", "validate": procstringT0},
                    "WriteTiers": {"default": ["RECO", "AOD", "DQM", "ALCARECO"],
                                   "type": makeList, "optional": False, "null": False},
                    "AlcaSkims": {"default": ["TkAlCosmics0T", "MuAlGlobalCosmics", "HcalCalHOCosmics"],
                                  "type": makeList, "optional": False, "null": False},
                    "PhysicsSkims": {"default": [], "type": makeList,
                                     "optional": True, "null": False},
                    "InitCommand": {"default": None, "optional": True, "null": True},
                    "EnvPath": {"default": None, "optional": True, "null": True},
                    "BinPath": {"default": None, "optional": True, "null": True},
                    "DoLogCollect": {"default": True, "type": strToBool,
                                     "optional": True, "null": False},
                    "SplittingAlgo": {"default": "EventAwareLumiBased", "null": False,
                                      "validate": lambda x: x in ["EventBased", "LumiBased",
                                                                  "EventAwareLumiBased", "FileBased"],
                                      "attr": "procJobSplitAlgo"},
                    "EventsPerJob": {"default": 500, "type": int, "validate": lambda x: x > 0,
                                     "null": False},
                    "SkimSplittingAlgo": {"default": "FileBased", "null": False,
                                          "validate": lambda x: x in ["EventBased", "LumiBased",
                                                                      "EventAwareLumiBased", "FileBased"],
                                          "attr": "skimJobSplitAlgo"},
                    "SkimEventsPerJob": {"default": 500, "type": int, "validate": lambda x: x > 0,
                                         "null": False},
                    "SkimLumisPerJob": {"default": 8, "type": int, "validate": lambda x: x > 0,
                                        "null": False},
                    "SkimFilesPerJob": {"default": 1, "type": int, "validate": lambda x: x > 0,
                                        "null": False},
                    }

        baseArgs.update(specArgs)
        DataProcessing.setDefaultArgumentsProperty(baseArgs)
        return baseArgs

    @staticmethod
    def getWorkloadAssignArgs():
        baseArgs = DataProcessing.getWorkloadAssignArgs()
        specArgs = {
            "Override": {"default": {"eos-lfn-prefix": "root://eoscms.cern.ch//eos/cms/store/logs/prod/recent/PromptReco"},
                         "type": dict},
            }
        baseArgs.update(specArgs)
        DataProcessing.setDefaultArgumentsProperty(baseArgs)
        return baseArgs



#!/usr/bin/env python
"""
_Unpacker_

Minimal WN tool to unpack the WMSandbox and start bootstrapping the job

Inputs are:

1. Task level Sandbox file
2. Job Package file
3. Job Package Index
4. Job Name is needed to generate an error report in case of failure

This will then do the following:
- create a job directory
- unpack the sandbox
- drop the package and index into the job area
- set everything up so that you can just add the job dir to the pythonpath and then call the runtime startup for the WMCore/WMRuntime stuff

"""
from __future__ import print_function

import getopt
import logging
import os
import sys
import tarfile
import traceback
import zipfile

options = {
    "sandbox=": "WMAGENT_SANDBOX",  # sandbox archive file
    "package=": "WMAGENT_PACKAGE",  # job package pickle file
    "index=": "WMAGENT_INDEX",  # index of job to be run
    "jobname=": "WMAGENT_JOBNAME",  # job name/id
}


def makeErrorReport(jobName, exitCode, message):
    """
    _makeErrorReport_

    Generate a Job Report indicating failure to bootstrap.
    Build as simple string to avoid dependencies on libraries required
    during bootstrap which may have failed.

    TODO: Include side & node information lookup
    TODO: FwkJobRep objects need refactor/cleanup for WMAgent
    TODO: Identify set of error codes

    """
    xml = "<xml>\n"
    xml += "<FrameworkJobReport Status=\"Failed\" Name=\"%s\">\n" % jobName
    xml += "<ExitCode Value=\"%s\"/>\n" % exitCode
    xml += "<FrameworkError Type = \"BootstrapError\">\n"
    xml += str(message)
    xml += "\n"
    xml += "</FrameworkError>\n"
    xml += "</FrameworkJobReport>\n"
    xml += "</xml>\n"
    with open("FrameworkJobReport.xml", 'w') as handle:
        handle.write(xml)


def createWorkArea(sandbox):
    """
    _createWorkArea_

    Create a job working area containing all the bits and pieces
    needed to bootstrap up and kickstart the job

    """
    currentDir = os.getcwd()
    jobDir = "%s/job" % currentDir
    if not os.path.exists(jobDir):
        os.makedirs(jobDir)
    if not os.path.exists(os.path.join(jobDir, 'StartupScript')):
        os.makedirs(os.path.join(jobDir, 'StartupScript'))

    with tarfile.open(sandbox, "r") as tfile:
        tfile.extractall(jobDir)

    # need to pull out the startup file from the zipball
    with zipfile.ZipFile(os.path.join(jobDir, 'WMCore.zip'), 'r') as zfile:
        startupScript = zfile.read('WMCore/WMRuntime/Startup.py')
        fd = os.open(os.path.join(jobDir, 'Startup.py'), os.O_CREAT | os.O_WRONLY)
        os.write(fd, startupScript)
        os.close(fd)

    logging.info("PYTHONPATH=%s", os.environ.get("PYTHONPATH"))

    return jobDir


def installPackage(jobArea, jobPackage, jobIndex):
    """
    _installPackage_

    Install the job package and index into the job directory so that
    it can be found on bootstrap

    """
    target = "%s/WMSandbox" % jobArea
    pkgTarget = "%s/JobPackage.pcl" % target
    # shutil.copy(jobPackage, pkgTarget)
    os.system("/bin/cp %s %s" % (jobPackage, pkgTarget))

    indexPy = "%s/JobIndex.py" % target
    with open(indexPy, 'w') as handle:
        handle.write("jobIndex = %s\n" % jobIndex)

    return


def runUnpacker(sandbox, package, jobIndex, jobname):
    """
    Run everything in the unpacker

    """

    try:
        jobArea = createWorkArea(sandbox)
        installPackage(jobArea, package, jobIndex)
        # sys.exit(0)
    except Exception as ex:
        msg = "Unable to create job area for bootstrap\n"
        msg += str(ex)
        msg += str(traceback.format_exc())
        makeErrorReport(jobname, 1, msg)
        logging.error(msg)
        sys.exit(1)


if __name__ == '__main__':

    try:
        opts, args = getopt.getopt(sys.argv[1:], "", options.keys())
    except getopt.GetoptError as ex:
        msg = "Error processing commandline args:\n"
        msg += str(ex)
        logging.error(msg)
        sys.exit(1)

    sandbox = os.environ.get('WMAGENT_SANDBOX', None)
    package = os.environ.get('WMAGENT_PACKAGE', None)
    jobIndex = os.environ.get('WMAGENT_INDEX', None)
    jobname = os.environ.get('WMAGENT_JOBNAME', None)
    for opt, arg in opts:
        if opt == "--sandbox":
            sandbox = arg
        if opt == "--package":
            package = arg
        if opt == "--index":
            jobIndex = arg
        if opt == "--jobname":
            jobname = arg

    if sandbox is None:
        msg = "No Sandbox provided"
        makeErrorReport(jobname, 1, msg)
        logging.error(msg)
        sys.exit(1)
    if package is None:
        msg = "No Job Package provided"
        makeErrorReport(jobname, 1, msg)
        logging.error(msg)
        sys.exit(1)
    if jobIndex is None:
        msg = "No Job Index provided"
        makeErrorReport(jobname, 1, msg)
        logging.error(msg)
        sys.exit(1)

    runUnpacker(sandbox=sandbox, package=package,
                jobIndex=jobIndex, jobname=jobname)



#!/usr/bin/env python
"""
_GetAllJobGroups_

Oracle implementation of Subscriptions.GetJobGroups
"""




from WMCore.WMBS.MySQL.Subscriptions.GetAllJobGroups import GetAllJobGroups as MySQLGetAllJobGroups

class GetAllJobGroups(MySQLGetAllJobGroups):
    """
    Identical to MySQL version

    """



#!/usr/bin/env python
"""
MySQL implementation of File.SetParentageByJob

Make the parentage link between a file and all the inputs of a given job
"""




import logging

from WMCore.Database.DBFormatter import DBFormatter

class SetParentageByJob(DBFormatter):
    """
    Make the parentage link between a file and all the inputs of a given job
    """


    sql = """INSERT INTO wmbs_file_parent (child, parent)
             SELECT DISTINCT wmbs_file_details.id, wmbs_job_assoc.fileid
             FROM wmbs_file_details, wmbs_job_assoc
             WHERE wmbs_job_assoc.job  = :jobid
             AND wmbs_file_details.lfn = :child
    """


    def execute(self, binds, conn = None, transaction = False):
        """
        Expect binds of form {'jobid', 'child'}

        """

        result = self.dbi.processData(self.sql, binds,
                         conn = conn, transaction = transaction)

        return



#!/usr/bin/env python
"""
_LoadFilesByWorkflow_

MySQL implementation of LoadFilesByWorkflow
"""




import logging

from WMComponent.DBS3Buffer.MySQL.LoadDBSFilesByDAS import LoadDBSFilesByDAS as MySQLLoadDBSFilesByDAS

class LoadFilesByWorkflow(MySQLLoadDBSFilesByDAS):
    fileInfoSQL = """SELECT files.id AS id, files.lfn AS lfn, files.filesize AS filesize,
                    files.events AS events,
                    files.status AS status,
                    dbsbuffer_algo.app_name AS app_name, dbsbuffer_algo.app_ver AS app_ver,
                    dbsbuffer_algo.app_fam AS app_fam, dbsbuffer_algo.pset_hash AS pset_hash,
                    dbsbuffer_algo.config_content, dbsbuffer_dataset.path AS dataset_path,
                    dbsbuffer_dataset.acquisition_era AS acquisition_era,
                    dbsbuffer_dataset.processing_ver AS processing_ver,
                    dbsbuffer_dataset.global_tag AS global_tag,
                    dbsbuffer_dataset.prep_id AS prep_id
             FROM dbsbuffer_file files
             INNER JOIN dbsbuffer_algo_dataset_assoc ON
               files.dataset_algo = dbsbuffer_algo_dataset_assoc.id
             INNER JOIN dbsbuffer_algo ON
               dbsbuffer_algo_dataset_assoc.algo_id = dbsbuffer_algo.id
             INNER JOIN dbsbuffer_dataset ON
               dbsbuffer_algo_dataset_assoc.dataset_id = dbsbuffer_dataset.id
             WHERE files.workflow = (SELECT id FROM dbsbuffer_workflow WHERE name = :workflow)
             ORDER BY files.id"""

    def execute(self, workflowName, conn = None, transaction = False):
        """
        This loads all files in a workflow.

        It has no defenses against loading files with unloaded parents.
        It depends on nobody but LoadDBSFilesByDAS putting files into a block
        """
        result   = self.dbi.processData(self.fileInfoSQL, {'workflow': workflowName},
                                        conn = conn, transaction = transaction)
        fileInfo = self.formatFileInfo(result)

        fileIDs  = [x['id'] for x in fileInfo]
        binds    = self.getBinds(fileIDs)

        if len(fileInfo) == 0:
            # Then we have no files for this workflow
            return []

        # Do locations
        result      = self.dbi.processData(self.getLocationSQL, binds, conn=conn, transaction=transaction)
        locInfo     = self.locInfo(result)
        fullResults = self.merge(fileInfo, locInfo)

        # Do checksums
        result      = self.dbi.processData(self.getChecksumSQL, binds, conn=conn, transaction=transaction)
        ckInfo      = self.ckInfo(result)
        fullResults = self.merge(fullResults, ckInfo)

        # Do runLumi
        result      = self.dbi.processData(self.getRunLumiSQL, binds, conn=conn, transaction=transaction)
        runInfo     = self.runInfo(result)
        fullResults = self.merge(fullResults, runInfo)

        # Do parents
        result      = self.dbi.processData(self.getParentLFNSQL, binds, conn=conn, transaction=transaction)
        parInfo     = self.parentInfo(result)
        fullResults = self.merge(fullResults, parInfo)

        return fullResults



"""
EmailAlert - send alerts via email

NOTICE:
This class does not work from kubernetes pods. The AlertManagerAPI class should be used for alerting from k8s.
More details at: https://github.com/dmwm/WMCore/issues/10234
"""

from __future__ import division
from builtins import str, object
import smtplib
import logging


class EmailAlert(object):
    """
    A simple class to send alerts via email
    """

    EMAIL_HEADER = "From: %s\r\nSubject: %s\r\nTo: %s\r\n\r\n"

    def __init__(self, configDict):
        self.serverName = configDict.get("smtpServer", "localhost")
        self.fromAddr = configDict.get("fromAddr", "noreply@cern.ch")
        self.toAddr = configDict.get("toAddr", "cms-service-production-admins@cern.ch")
        if not isinstance(self.toAddr, (list, set)):
            self.toAddr = [self.toAddr]

    def send(self, subject, message):
        """
        Send an email
        :param subject: Email subject
        :param message: Email body
        """
        msg = self.EMAIL_HEADER % (self.fromAddr, subject, ", ".join(self.toAddr))
        msg += message

        try:
            smtp = smtplib.SMTP(self.serverName)
            smtp.sendmail(self.fromAddr, self.toAddr, msg)
        except Exception as ex:
            logging.exception("Error sending alert email.\nDetails: %s", str(ex))

        try:
            # clean up smtp connection
            smtp.quit()
        except UnboundLocalError:
            # it means our client failed connecting to the SMTP server
            pass




#!/usr/bin/python
"""
_Create_

Class for creating MySQL specific schema for resource control.
"""

import threading

from WMCore.Database.DBCreator import DBCreator


class Create(DBCreator):
    """
    _Create_

    Class for creating MySQL specific schema for resource control.
    """

    def __init__(self, logger=None, dbi=None, params=None):
        myThread = threading.currentThread()
        DBCreator.__init__(self, myThread.logger, myThread.dbi)
        self.create = {}
        self.constraints = {}

        self.create["rc_threshold"] = """
        CREATE TABLE rc_threshold(
            site_id       INTEGER NOT NULL,
            sub_type_id   INTEGER NOT NULL,
            pending_slots INTEGER NOT NULL,
            max_slots     INTEGER NOT NULL,
            FOREIGN KEY (site_id) REFERENCES wmbs_location(id) ON DELETE CASCADE,
            FOREIGN KEY (sub_type_id) REFERENCES wmbs_sub_types(id) ON DELETE CASCADE
            ) ENGINE=InnoDB ROW_FORMAT=DYNAMIC"""

        return



#!/usr/bin/env python
"""
_LoadOutput_

Oracle implementation of Workflow.LoadOutput
"""

from WMCore.WMBS.MySQL.Workflow.LoadOutput import LoadOutput as LoadOutputMySQL

class LoadOutput(LoadOutputMySQL):
    pass



#!/usr/bin/env python
"""
_GetType_

MySQL implementation of Jobs.GetType
"""




from WMCore.Database.DBFormatter import DBFormatter

class GetType(DBFormatter):
    """
    _GetType_

    Given a job ID, get the type of job from the subscription.
    """
    sql = """SELECT wmbs_job.id AS id, wmbs_sub_types.name AS type
             FROM wmbs_sub_types
               INNER JOIN wmbs_subscription ON
                 wmbs_sub_types.id = wmbs_subscription.subtype
               INNER JOIN wmbs_jobgroup ON
                 wmbs_subscription.id = wmbs_jobgroup.subscription
               INNER JOIN wmbs_job ON
                 wmbs_jobgroup.id = wmbs_job.jobgroup
             WHERE wmbs_job.id = :jobid"""

    def execute(self, jobID, conn = None, transaction = False):
        isList = isinstance(jobID, list)
        if isList:
            binds = []
            for job in jobID:
                binds.append({"jobid": job})
        else:
            binds = {"jobid": jobID}
        result = self.dbi.processData(self.sql, binds, conn = conn,
                                      transaction = transaction)
        if isList:
            return self.formatDict(result)
        else:
            return self.formatOneDict(result)["type"]



#!/usr/bin/env python
"""
_CountWorkflowBySpec_

MySQL implementation of Workflow.CountWorkflowBySpec
"""

from WMCore.Database.DBFormatter import DBFormatter

class CountWorkflowBySpec(DBFormatter):
    sql = """SELECT COUNT(*) AS workflows FROM wmbs_workflow WHERE spec = :spec"""

    def execute(self, spec, conn = None, transaction = False):
        result = self.dbi.processData(self.sql, {"spec": spec},
                                      conn = conn, transaction = transaction)
        return self.formatDict(result)[0].get('workflows', 0)



#!/usr/bin/env python
"""
Oracle implementation of File.GetLocation
"""




#This has been modified for Oracle

from WMComponent.DBS3Buffer.MySQL.DBSBufferFiles.GetLocation import GetLocation as MySQLGetLocation

class GetLocation(MySQLGetLocation):
    sql = """select pnn from dbsbuffer_location
                where id in (select location from dbsbuffer_file_location
                    where filename in (select id from dbsbuffer_file where lfn=:lfn))"""



"""
_IsCompleteOnRun_
MySQL implementation of Subscriptions.IsCompleteOnRun

Checks all files in the given subscription and given run are completed.
"""
__all__ = []



from WMCore.Database.DBFormatter import DBFormatter

class IsCompleteOnRun(DBFormatter):
    """
    _IsCompleteOnRun_

    Return number of files for available, complete, failed status
    for a given run and a given subscription.

    TODO: can use left outer join to check the completeness.
    Not sure join is more expensive than multiple select with count
    """
    sql = """SELECT count(*) FROM wmbs_fileset_files wff
                INNER JOIN wmbs_subscription ws ON (ws.fileset = wff.fileset)
                INNER JOIN wmbs_file_runlumi_map wrm ON (wrm.fileid = wff.fileid)
                LEFT OUTER JOIN wmbs_sub_files_failed wf ON (wf.fileid = wff.fileid AND wf.subscription = ws.id)
                LEFT OUTER JOIN wmbs_sub_files_complete wc ON (wc.fileid = wff.fileid AND wc.subscription = ws.id)
                WHERE
                 wf.fileid IS NULL AND wc.fileid IS NULL AND
                 ws.id = :subID AND wrm.run = :runID
          """

    def format(self, result):
        result = DBFormatter.format(self, result)
        return result[0][0]

    def execute(self, subID, runID, conn = None, transaction = False):
        binds = self.getBinds(subID=subID, runID=runID)
        result = self.dbi.processData(self.sql, binds, conn = conn,
                                      transaction = transaction)
        return self.format(result)



#!/usr/bin/env python
# pylint: disable=E1101

"""
Harness class that wraps standard functionality used in all daemon
components including:

(1) Instantiating message service, trigger and other database classes
including session objects and workflow entities.

(2) Subscribing to default messages (e.g. logging and stopping server)

(3) Providing a method to test the component without a message service

(4) Instatiate the log master

(5) Provides basic uniform log information (useful for operators)

(6) Method to publish monitoring information

"""
from __future__ import print_function
from builtins import object

import logging
import os
import sys
import threading
import time
import traceback
from logging.handlers import RotatingFileHandler

from WMCore import WMLogging
from WMCore.Agent.ConfigDBMap import ConfigDBMap
from WMCore.Agent.Daemon.Create import createDaemon
from WMCore.Agent.HeartbeatAPI import HeartbeatAPI
from WMCore.Database.DBFactory import DBFactory
from WMCore.Database.Transaction import Transaction
from WMCore.WMException import WMException
from WMCore.WMExceptions import WMEXCEPTION
from WMCore.WorkerThreads.WorkerThreadManager import WorkerThreadManager


class HarnessException(WMException):
    """
    _HarnessException_

    Well, it has the harness errors in it.
    Otherwise, it's just part of WMException.
    """


class Harness(object):
    """
    Harness class that wraps standard functionality used in all daemon
    components
    """

    def __init__(self, config, compName=None):
        """
        init

        The constructor is empty as we have an initalization method
        that can be called inside new threads (we use thread local attributes
        at startup.

        Default intialization of the harness including setting some diagnostic
        messages
        """
        self.config = config

        # component name is always the class name of child class
        if not compName:
            compName = self.__class__.__name__

        if compName not in (self.config.listComponents_() + self.config.listWebapps_()):
            raise WMException(WMEXCEPTION['WMCORE-8'] + compName, 'WMCORE-8')
        if not hasattr(self.config, "Agent"):
            self.config.section_("Agent")

        self.config.Agent.componentName = compName
        compSect = getattr(self.config, compName, None)
        if compSect is None:
            # Then we have a major problem - there's no section with this name
            logging.error("Could not find section %s in config", compName)
            logging.error("We are returning, and hoping you know what you're doing!")
            logging.debug("Config: %s", self.config)
            return
        # check if componentDir is set if not assign.
        if getattr(compSect, 'componentDir', None) is None:
            if not hasattr(self.config, "General"):
                # Don't do anything.  Assume the user knows what they are doing.
                logging.error("Missing componentDir and General section in config")
                logging.error("Going to trust you to know what you're doing.")
                return

            compSect.componentDir = os.path.join(self.config.General.workDir,
                                                 'Components',
                                                 self.config.Agent.componentName)
        # we have name and location of the log files. Now make sure there
        # is a directory.
        try:
            if not os.path.isdir(compSect.componentDir):
                os.makedirs(compSect.componentDir)
        except Exception as ex:
            logging.error("Encountered exception while making componentDirs: %s", str(ex))
            logging.error("Ignoring")

        self.threadManagerName = ''
        self.heartbeatAPI = None
        self.messages = {}
        self.logMsg = {}

        return

    def initInThread(self):
        """
        Default intialization of the harness including setting some diagnostic
        messages. This method is called when we call 'prepareToStart'
        """
        try:
            self.messages = {}

            compName = self.config.Agent.componentName
            compSect = getattr(self.config, compName, None)
            if not hasattr(compSect, "logFile"):
                if not getattr(compSect, 'componentDir', None):
                    errorMessage = "No componentDir for log entries found!\n"
                    errorMessage += "Harness cannot run without componentDir.\n"
                    logging.error(errorMessage)
                    raise HarnessException(errorMessage)
                compSect.logFile = os.path.join(compSect.componentDir, "ComponentLog")
            print('Log file is: ' + compSect.logFile)
            logHandler = RotatingFileHandler(compSect.logFile,
                                             "a", 1000000000, 3)
            logMsgFormat = getattr(compSect, "logMsgFormat",
                                   "%(asctime)s:%(thread)d:%(levelname)s:%(module)s:%(message)s")
            logFormatter = \
                logging.Formatter(logMsgFormat)
            logHandler.setFormatter(logFormatter)
            logLevelName = getattr(compSect, 'logLevel', 'INFO')
            logLevel = getattr(logging, logLevelName)
            logging.getLogger().addHandler(logHandler)
            logging.getLogger().setLevel(logLevel)
            self.logMsg = {'DEBUG': logging.DEBUG,
                           'ERROR': logging.ERROR,
                           'NOTSET': logging.NOTSET,
                           'CRITICAL': logging.CRITICAL,
                           'WARNING': logging.WARNING,
                           'INFO': logging.INFO,
                           'SQLDEBUG': logging.SQLDEBUG}
            if hasattr(compSect, "logLevel") and compSect.logLevel in self.logMsg:
                logging.getLogger().setLevel(self.logMsg[compSect.logLevel])
            WMLogging.sqldebug("wmcore level debug:")

            # If not previously set, force wmcore cache to current path
            if not os.environ.get('WMCORE_CACHE_DIR'):
                os.environ['WMCORE_CACHE_DIR'] = os.path.join(compSect.componentDir, '.wmcore_cache')

            logging.info(">>>Starting: " + compName + '<<<')
            # check which backend to use: MySQL, Oracle, etc... for core
            # services.
            # we recognize there can be more than one database.
            # be we offer a default database that is used for core services.
            logging.info(">>>Initializing default database")
            logging.info(">>>Check if connection is through socket")
            myThread = threading.currentThread()
            myThread.logger = logging.getLogger()
            logging.info(">>>Setting config for thread: ")
            myThread.config = self.config

            logging.info(">>>Building database connection string")
            # check if there is a premade string if not build it yourself.
            dbConfig = ConfigDBMap(self.config)
            dbStr = dbConfig.getDBUrl()
            options = dbConfig.getOption()
            # we only want one DBFactory per database so we will need to
            # to pass this on in case we are using threads.
            myThread.dbFactory = DBFactory(myThread.logger, dbStr, options)

            myThread.sql_transaction = True
            if myThread.dbFactory.engine:

                myThread.dbi = myThread.dbFactory.connect()
                myThread.transaction = Transaction(myThread.dbi)

            else:

                myThread.dbi = myThread.config.CoreDatabase.connectUrl
                myThread.sql_transaction = False

            # Attach a worker manager object to the main thread
            if not hasattr(myThread, 'workerThreadManager'):
                myThread.workerThreadManager = WorkerThreadManager(self)
            else:
                myThread.workerThreadManager.terminateSlaves.clear()
            myThread.workerThreadManager.pauseWorkers()

            logging.info(">>>Initialize transaction dictionary")

            (connectDialect, dummy) = dbStr.split(":", 1)

            if connectDialect.lower() == 'mysql':
                myThread.dialect = 'MySQL'
            elif connectDialect.lower() == 'oracle':
                myThread.dialect = 'Oracle'

            logging.info("Harness part constructor finished")
        except Exception as ex:
            logging.critical("Problem instantiating " + str(ex))
            logging.error("Traceback: %s", str(traceback.format_exc()))
            raise

    def preInitialization(self):
        """
        _preInitialization_

        returns: nothing

        method that can be overloaded and will be called before the
        start component is called. (enables you to set message->handler
        mappings). You use the self.message dictionary of the base class
        to define the mappings.

        """
        pass

    def postInitialization(self):
        """
        _postInitialization_

        returns: nothing

        method that can be overloaded and will be called after the start
        component does the standard initialization, but before the wait
        (enables you to publish events when starting up)

        Define actions you want to execute before the actual message
        handling starts. E.g.: publishing some messages, or removing
        messages.

        """
        pass

    def logState(self):
        """
        _logState_

        returns: string

        method that can be overloaded to log additional state information
        (should return atring)
        """
        msg = 'No additional state information for ' + \
              self.config.Agent.componentName
        return msg

    def publishItem(self, items):
        """
        _publishItem_

        returns: nothing

        A method that publishes a (dictionary) set or 1 item
        to a monitoring service.
        """
        # FIXME: do we need this method. If so we need to agree
        # FIXME: on some default monitoring publication mechanism.
        pass

    def __call__(self, event, payload):
        """
        Once upon a time this was for doing the handling of diagnostic messages

        With the test-deprecating of the MsgService based diagnostics, we've basically
        scratched this.

        I'm leaving this in so at least the framework is still there

        -mnorman
        """
        return

    def initialization(self):
        """
        _initialization__

        Used the handle initializing the MsgService.  The MsgService
        is no longer used.

        Removed but not deleted, since all sorts of things call it
        """
        return

    def prepareToStart(self):
        """
        _prepareToStart_

        returns: Nothing

        Starts the initialization procedure. It is mainly an aggregation method
        so it can easily used in tests.
        """
        self.state = 'initialize'
        self.initInThread()
        # note: every component gets a (unique) name:
        # self.config.Agent.componentName
        logging.info(">>>Registering Component - %s", self.config.Agent.componentName)

        if getattr(self.config.Agent, "useHeartbeat", True):
            self.heartbeatAPI = HeartbeatAPI(self.config.Agent.componentName)
            self.heartbeatAPI.registerComponent()

        logging.info('>>>Starting initialization')

        logging.info('>>>Setting default transaction')
        myThread = threading.currentThread()

        self.preInitialization()

        if myThread.sql_transaction:
            myThread.transaction.begin()

        self.initialization()
        self.postInitialization()

        if myThread.sql_transaction:
            myThread.transaction.commit()

        logging.info('>>>Committing default transaction')

        logging.info(">>>Starting worker threads")
        myThread.workerThreadManager.resumeWorkers()

        logging.info(">>>Initialization finished!\n")
        # wait for messages
        self.state = 'active'

    def prepareToStop(self, wait=False, stopPayload=""):
        """
        _stopComponent

        Stops the component, including all worker threads. Allows call from
        test framework
        """
        # Stop all worker threads
        logging.info(">>>Terminating worker threads")
        myThread = threading.currentThread()
        try:
            myThread.workerThreadManager.terminateWorkers()
        except Exception:
            # We may not have a thread manager
            pass

        if wait:
            logging.info(">>>Shut down of component while waiting for threads to finish")
            # check if nr of threads is specified.
            activeThreads = 1
            if stopPayload != "":
                activeThreads = int(stopPayload)
                if activeThreads < 1:
                    activeThreads = 1
            while threading.activeCount() > activeThreads:
                logging.info('>>>Currently ' + str(threading.activeCount()) + ' threads active')
                logging.info('>>>Waiting for less than ' + str(activeThreads) + ' to be active')
                time.sleep(5)

    def handleMessage(self, type='', payload=''):
        """
        __handleMessage_

        Formerly used to handle messages - now non-functional
        Left here in case someone else is using it (i.e. PilotManager)
        """
        return

    def startDaemon(self, keepParent=False, compName=None):
        """
        Same result as start component, except that the comopnent
        is started as a daemon, after which you can close your xterm
        and the process will still run.

        The keepParent option enables us to keep the parent process
        which is used during testing,
        """
        msg = "Starting %s as a daemon " % self.config.Agent.componentName
        print(msg)
        if not compName:
            compName = self.__class__.__name__
        compSect = getattr(self.config, compName, None)
        msg = "Log will be in %s " % compSect.componentDir
        print(msg)
        # put the daemon config file in the work dir of this component.
        # FIXME: this file will be replaced by a database table.
        compSect = getattr(self.config, self.config.Agent.componentName, None)
        pid = createDaemon(compSect.componentDir, keepParent)
        # if this is not the parent start the component
        if pid == 0:
            self.startComponent()
            # if this is the parent return control to the testing environment.

    def startComponent(self):
        """
        _startComponent_

        returns: Nothing

        Start up the component, performs initialization and waits indefinitely
        Calling this method results in the application
        running in the xterm (not in daemon mode)

        """
        myThread = threading.currentThread()
        try:
            msg = 'None'
            self.prepareToStart()
            while True:
                time.sleep(360)

        except Exception as ex:
            if self.state == 'initialize':
                errormsg = """PostMortem: choked when initializing with error: %s\n""" % (str(ex))
                stackTrace = traceback.format_tb(sys.exc_info()[2], None)
                for stackFrame in stackTrace:
                    errormsg += stackFrame
            else:
                errormsg = ""
                stackTrace = traceback.format_tb(sys.exc_info()[2], None)
                for stackFrame in stackTrace:
                    errormsg += stackFrame
                logging.error(errormsg)
                logging.error(">>>Fatal Error, Preparing to Rollback Transaction")
                if getattr(myThread, 'transaction', None) is not None:
                    myThread.transaction.rollback()
                self.prepareToStop(False)
                errormsg = """
PostMortem: choked while handling messages  with error: %s
while trying to handle msg: %s
                """ % (str(ex), str(msg))
            print(errormsg)
            logging.critical(errormsg)
            raise
        logging.info("System shutdown complete!")
        # this is to ensure exiting when in daemon mode.
        sys.exit()

    def __str__(self):
        """

        return: string

        String representation of the status of this component.
        """

        msg = 'Status of this component : \n'
        msg += '\n'
        msg += '>>Event Subscriptions --> Handlers<<\n'
        msg += '------------------------------------\n'
        for message in self.messages:
            msg += message + '-->' + str(self.messages[message]) + '\n'
        msg += '\n'
        msg += '\n'
        msg += '>>Parameters --> Values<<\n'
        msg += '-------------------------\n'
        msg += str(self.config)
        additionalMsg = self.logState()
        if additionalMsg != '':
            msg += '\n'
            msg += 'Additional state information\n'
            msg += '----------------------------\n'
            msg += '\n'
            msg += str(additionalMsg)
            msg += '\n'
        return msg



#!/usr/bin/env python

"""
_GetLocationBulk_

Oracle implementation of Files.GetLocationBulk
"""


from WMCore.WMBS.MySQL.Files.GetLocationBulk import GetLocationBulk \
     as GetLocationBulkMySQL

class GetLocationBulk(GetLocationBulkMySQL):
    """
    _GetLocationBulk_

    Oracle specific: file is reserved word

    """

    pass



#!/usr/bin/env python
"""
_GetFinishedTasks_

MySQL implementation of Workflow.GetFinishedTasks

Created on Nov 7, 2012

@author: dballest
"""

from WMCore.Database.DBFormatter import DBFormatter

class GetFinishedTasks(DBFormatter):
    """
    _GetFinishedTasks_

    Returns those tasks where the associated subscriptions
    are marked as finished
    """

    sql = """SELECT wmbs_workflow.name, wmbs_workflow.task,
                    wmbs_workflow.spec
             FROM wmbs_workflow
             LEFT OUTER JOIN wmbs_subscription ON
                 wmbs_subscription.workflow = wmbs_workflow.id AND
                 wmbs_subscription.finished = 0
            GROUP BY wmbs_workflow.name, wmbs_workflow.task,
                     wmbs_workflow.spec
            HAVING COUNT(wmbs_subscription.id) = 0
          """

    def execute(self, conn = None, transaction = False):
        """
        _execute_

        Returns a list of dictionaries with the following structure:
        [ {'name' : <requestName>, 'task' : <taskPath>}, ...]
        """
        result = self.dbi.processData(self.sql,
                                        conn = conn, transaction = transaction)
        return self.formatDict(result)



"""
_CreateWMBS_

Base class for creating the WMBS database.
"""

import threading

from WMCore.Database.DBCreator import DBCreator
from WMCore.JobStateMachine.Transitions import Transitions
from WMCore.WMException import WMException


class CreateWMBSBase(DBCreator):
    def __init__(self, logger=None, dbi=None, params=None):
        """
        _init_

        Call the DBCreator constructor and create the list of required tables.
        """
        myThread = threading.currentThread()

        if logger is None:
            logger = myThread.logger
        if dbi is None:
            dbi = myThread.dbi

        tablespaceIndex = ""
        if params:
            if "tablespace_index" in params:
                tablespaceIndex = "USING INDEX TABLESPACE %s" % params["tablespace_index"]

        DBCreator.__init__(self, logger, dbi)

        self.requiredTables = ["01wmbs_fileset",
                               "02wmbs_file_details",
                               "03wmbs_fileset_files",
                               "04wmbs_file_parent",
                               "05wmbs_file_runlumi_map",
                               "05wmbs_location_state",
                               "06wmbs_location",
                               "06wmbs_pnns",
                               "07wmbs_location_pnns",
                               "07wmbs_file_location",
                               "07wmbs_users",
                               "07wmbs_workflow",
                               "08wmbs_sub_types",
                               "08wmbs_workflow_output",
                               "09wmbs_subscription",
                               "10wmbs_subscription_validation",
                               "10wmbs_sub_files_acquired",
                               "10wmbs_sub_files_available",
                               "11wmbs_sub_files_failed",
                               "12wmbs_sub_files_complete",
                               "13wmbs_jobgroup",
                               "14wmbs_job_state",
                               "15wmbs_job",
                               "16wmbs_job_assoc",
                               "17wmbs_job_mask",
                               "18wmbs_checksum_type",
                               "19wmbs_file_checksums",
                               "21wmbs_workunit",
                               "22wmbs_job_workunit_assoc",
                               "23wmbs_frl_workunit_assoc",
                               ]

        self.create["01wmbs_fileset"] = \
            """CREATE TABLE wmbs_fileset (
               id          INTEGER      PRIMARY KEY AUTO_INCREMENT,
               name        VARCHAR(1250) NOT NULL,
               open        INT(1)       NOT NULL DEFAULT 0,
               last_update INTEGER      NOT NULL,
               UNIQUE (name))"""

        self.create["02wmbs_file_details"] = \
            """CREATE TABLE wmbs_file_details (
               id           INTEGER      PRIMARY KEY AUTO_INCREMENT,
               lfn          VARCHAR(1250) NOT NULL,
               filesize     BIGINT,
               events       BIGINT UNSIGNED,
               first_event  BIGINT       UNSIGNED NOT NULL DEFAULT 0,
               merged       INT(1)       NOT NULL DEFAULT 0,
               UNIQUE (lfn))"""

        self.create["03wmbs_fileset_files"] = \
            """CREATE TABLE wmbs_fileset_files (
               fileid      INTEGER   NOT NULL,
               fileset     INTEGER   NOT NULL,
               insert_time INTEGER   NOT NULL,
               UNIQUE (fileid, fileset),
               FOREIGN KEY(fileset) references wmbs_fileset(id))"""

        self.create["04wmbs_file_parent"] = \
            """CREATE TABLE wmbs_file_parent (
               child  INTEGER NOT NULL,
               parent INTEGER NOT NULL,
               FOREIGN KEY (child)  references wmbs_file_details(id)
                 ON DELETE CASCADE,
               FOREIGN KEY (parent) references wmbs_file_details(id),
               UNIQUE(child, parent))"""

        self.create["05wmbs_file_runlumi_map"] = \
            """CREATE TABLE wmbs_file_runlumi_map (
               fileid  INTEGER NOT NULL,
               run     INTEGER NOT NULL,
               lumi    INTEGER NOT NULL,
               num_events  BIGINT UNSIGNED,
               PRIMARY KEY (fileid, run, lumi),
               FOREIGN KEY (fileid) references wmbs_file_details(id)
                 ON DELETE CASCADE)"""

        self.create["05wmbs_location_state"] = \
            """CREATE TABLE wmbs_location_state (
               id   INTEGER PRIMARY KEY AUTO_INCREMENT,
               name VARCHAR(100) NOT NULL)"""

        self.create["06wmbs_location"] = \
            """CREATE TABLE wmbs_location (
               id          INTEGER      PRIMARY KEY AUTO_INCREMENT,
               site_name   VARCHAR(255) NOT NULL,
               state       INTEGER NOT NULL,
               cms_name    VARCHAR(255),
               ce_name     VARCHAR(255),
               running_slots   INTEGER,
               pending_slots   INTEGER,
               plugin      VARCHAR(255),
               state_time  INTEGER DEFAULT 0,
               UNIQUE(site_name),
               FOREIGN KEY (state) REFERENCES wmbs_location_state(id))"""

        self.create["06wmbs_pnns"] = \
            """CREATE TABLE wmbs_pnns (
                 id   INTEGER      PRIMARY KEY AUTO_INCREMENT,
                 pnn    VARCHAR(255),
                 UNIQUE(pnn))"""

        self.create["07wmbs_location_pnns"] = \
            """CREATE TABLE wmbs_location_pnns (
                 location   INTEGER,
                 pnn    INTEGER,
                 UNIQUE(location, pnn),
                 FOREIGN KEY (location) REFERENCES wmbs_location(id)
                   ON DELETE CASCADE,
                 FOREIGN KEY (pnn) REFERENCES wmbs_pnns(id)
                   ON DELETE CASCADE)"""

        self.create["07wmbs_users"] = \
            """CREATE TABLE wmbs_users (
               id        INTEGER      PRIMARY KEY AUTO_INCREMENT,
               cert_dn   VARCHAR(255) NOT NULL,
               name_hn   VARCHAR(255),
               owner     VARCHAR(255),
               grp       VARCHAR(255),
               group_name     VARCHAR(255),
               role_name      VARCHAR(255),
               UNIQUE(cert_dn, group_name, role_name))"""

        self.create["07wmbs_file_location"] = \
            """CREATE TABLE wmbs_file_location (
               fileid   INTEGER NOT NULL,
               pnn      INTEGER NOT NULL,
               UNIQUE(fileid, pnn),
               FOREIGN KEY(fileid)   REFERENCES wmbs_file_details(id)
                 ON DELETE CASCADE,
               FOREIGN KEY(pnn) REFERENCES wmbs_pnns(id)
                 ON DELETE CASCADE)"""

        self.create["07wmbs_workflow"] = \
            """CREATE TABLE wmbs_workflow (
               id           INTEGER          PRIMARY KEY AUTO_INCREMENT,
               spec         VARCHAR(700)     NOT NULL,
               name         VARCHAR(255)     NOT NULL,
               task         VARCHAR(1250)     NOT NULL,
               type         VARCHAR(255),
               owner        INTEGER          NOT NULL,
               alt_fs_close INT(1)           NOT NULL,
               injected     INT(1)           DEFAULT 0,
               priority     INTEGER UNSIGNED DEFAULT 0,
               FOREIGN KEY (owner)
               REFERENCES wmbs_users(id) ON DELETE CASCADE) """

        self.indexes["03_pk_wmbs_workflow"] = \
            """ALTER TABLE wmbs_workflow ADD
                 (CONSTRAINT wmbs_workflow_unique UNIQUE (name, spec, task))"""

        self.create["08wmbs_workflow_output"] = \
            """CREATE TABLE wmbs_workflow_output (
               workflow_id           INTEGER NOT NULL,
               output_identifier     VARCHAR(255) NOT NULL,
               output_fileset        INTEGER NOT NULL,
               merged_output_fileset INTEGER,
               FOREIGN KEY(workflow_id)  REFERENCES wmbs_workflow(id)
                 ON DELETE CASCADE,
               FOREIGN KEY(output_fileset)  REFERENCES wmbs_fileset(id)
                 ON DELETE CASCADE,
               FOREIGN KEY(merged_output_fileset)  REFERENCES wmbs_fileset(id)
                 ON DELETE CASCADE)
               """

        self.create["08wmbs_sub_types"] = \
            """CREATE TABLE wmbs_sub_types (
                 id   INTEGER      PRIMARY KEY AUTO_INCREMENT,
                 name VARCHAR(255) NOT NULL,
                 priority INTEGER DEFAULT 0,
                 UNIQUE(name))"""

        self.create["09wmbs_subscription"] = \
            """CREATE TABLE wmbs_subscription (
               id          INTEGER      PRIMARY KEY AUTO_INCREMENT,
               fileset     INTEGER      NOT NULL,
               workflow    INTEGER      NOT NULL,
               split_algo  VARCHAR(255) NOT NULL,
               subtype     INTEGER      NOT NULL,
               last_update INTEGER      NOT NULL,
               finished    INT(1)       DEFAULT 0,
               FOREIGN KEY(fileset)
               REFERENCES wmbs_fileset(id) ON DELETE CASCADE,
               FOREIGN KEY(workflow)
               REFERENCES wmbs_workflow(id) ON DELETE CASCADE,
               FOREIGN KEY(subtype)
               REFERENCES wmbs_sub_types(id) ON DELETE CASCADE)"""

        self.create["10wmbs_subscription_validation"] = \
            """CREATE TABLE wmbs_subscription_validation (
               subscription_id INTEGER NOT NULL,
               location_id     INTEGER NOT NULL,
               valid           INTEGER,
               UNIQUE (subscription_id, location_id),
               FOREIGN KEY(subscription_id) REFERENCES wmbs_subscription(id)
                 ON DELETE CASCADE,
               FOREIGN KEY(location_id) REFERENCES wmbs_location(id)
                 ON DELETE CASCADE)"""

        self.create["10wmbs_sub_files_acquired"] = \
            """CREATE TABLE wmbs_sub_files_acquired (
               subscription INTEGER NOT NULL,
               fileid       INTEGER NOT NULL,
               PRIMARY KEY (subscription, fileid),
               FOREIGN KEY (subscription) REFERENCES wmbs_subscription(id)
                 ON DELETE CASCADE,
               FOREIGN KEY (fileid)       REFERENCES wmbs_file_details(id)
                 ON DELETE CASCADE)
               """

        self.create["10wmbs_sub_files_available"] = \
            """CREATE TABLE wmbs_sub_files_available (
               subscription INTEGER NOT NULL,
               fileid       INTEGER NOT NULL,
               PRIMARY KEY (subscription, fileid),
               FOREIGN KEY (subscription) REFERENCES wmbs_subscription(id)
                 ON DELETE CASCADE,
               FOREIGN KEY (fileid)       REFERENCES wmbs_file_details(id)
                 ON DELETE CASCADE)
               """

        self.create["11wmbs_sub_files_failed"] = \
            """CREATE TABLE wmbs_sub_files_failed (
               subscription INTEGER NOT NULL,
               fileid       INTEGER NOT NULL,
               PRIMARY KEY (subscription, fileid),
               FOREIGN KEY (subscription) REFERENCES wmbs_subscription(id)
                 ON DELETE CASCADE,
               FOREIGN KEY (fileid)       REFERENCES wmbs_file_details(id)
                 ON DELETE CASCADE)"""

        self.create["12wmbs_sub_files_complete"] = \
            """CREATE TABLE wmbs_sub_files_complete (
               subscription INTEGER NOT NULL,
               fileid       INTEGER NOT NULL,
               PRIMARY KEY (subscription, fileid),
               FOREIGN KEY (subscription) REFERENCES wmbs_subscription(id)
                 ON DELETE CASCADE,
               FOREIGN KEY (fileid)       REFERENCES wmbs_file_details(id)
                 ON DELETE CASCADE)"""

        self.create["13wmbs_jobgroup"] = \
            """CREATE TABLE wmbs_jobgroup (
               id           INTEGER      PRIMARY KEY AUTO_INCREMENT,
               subscription INTEGER      NOT NULL,
               guid         VARCHAR(255),
               output       INTEGER,
               last_update  INTEGER      NOT NULL,
               location     INTEGER,
               UNIQUE(guid),
               FOREIGN KEY (subscription) REFERENCES wmbs_subscription(id)
                 ON DELETE CASCADE,
               FOREIGN KEY (output) REFERENCES wmbs_fileset(id)
                      ON DELETE CASCADE)"""

        self.create["14wmbs_job_state"] = \
            """CREATE TABLE wmbs_job_state (
               id        INTEGER       PRIMARY KEY AUTO_INCREMENT,
               name      VARCHAR(100))"""

        self.create["15wmbs_job"] = \
            """CREATE TABLE wmbs_job (
               id           INTEGER       PRIMARY KEY AUTO_INCREMENT,
               jobgroup     INTEGER       NOT NULL,
               name         VARCHAR(255),
               state        INTEGER       NOT NULL,
               state_time   INTEGER       NOT NULL,
               retry_count  INTEGER       DEFAULT 0,
               couch_record VARCHAR(255),
               location     INTEGER,
               outcome      INTEGER       DEFAULT 0,
               cache_dir    VARCHAR(1250)  DEFAULT 'None',
               fwjr_path    VARCHAR(1250),
               FOREIGN KEY (jobgroup)
               REFERENCES wmbs_jobgroup(id) ON DELETE CASCADE,
               FOREIGN KEY (state) REFERENCES wmbs_job_state(id),
               FOREIGN KEY (location) REFERENCES wmbs_location(id))"""

        self.indexes["03_pk_wmbs_job"] = \
            """ALTER TABLE wmbs_job ADD
                 (CONSTRAINT wmbs_job_unique UNIQUE (name, cache_dir, fwjr_path))"""

        self.create["16wmbs_job_assoc"] = \
            """CREATE TABLE wmbs_job_assoc (
               job    INTEGER NOT NULL,
               fileid INTEGER NOT NULL,
               FOREIGN KEY (job)  REFERENCES wmbs_job(id)
                 ON DELETE CASCADE,
               FOREIGN KEY (fileid) REFERENCES wmbs_file_details(id)
                 ON DELETE CASCADE)"""

        self.create["17wmbs_job_mask"] = \
            """CREATE TABLE wmbs_job_mask (
                job           INTEGER     NOT NULL,
                FirstEvent    BIGINT,
                LastEvent     BIGINT,
                FirstLumi     INTEGER,
                LastLumi      INTEGER,
                FirstRun      INTEGER,
                LastRun       INTEGER,
                inclusivemask BOOLEAN DEFAULT TRUE,
                FOREIGN KEY (job)       REFERENCES wmbs_job(id)
                  ON DELETE CASCADE)"""

        self.create["18wmbs_checksum_type"] = \
            """CREATE TABLE wmbs_checksum_type (
                id            INTEGER      PRIMARY KEY AUTO_INCREMENT,
                type          VARCHAR(255) )
                """

        self.create["19wmbs_file_checksums"] = \
            """CREATE TABLE wmbs_file_checksums (
                fileid        INTEGER,
                typeid        INTEGER,
                cksum         VARCHAR(100),
                UNIQUE (fileid, typeid),
                FOREIGN KEY (typeid) REFERENCES wmbs_checksum_type(id)
                  ON DELETE CASCADE,
                FOREIGN KEY (fileid) REFERENCES wmbs_file_details(id)
                  ON DELETE CASCADE)"""

        # Workunit table for tracking individual lumis
        self.create["21wmbs_workunit"] = (
            'CREATE TABLE wmbs_workunit('
            ' id INTEGER PRIMARY KEY AUTO_INCREMENT,'
            ' taskid INTEGER NOT NULL,'
            ' retry_count INTEGER DEFAULT 0,'
            ' last_unit_count INTEGER NOT NULL,'
            ' last_submit_time INTEGER NOT NULL,'
            ' status INT(1) DEFAULT 0,'
            ' FOREIGN KEY(taskid) REFERENCES wmbs_workflow(id) ON DELETE CASCADE)'
        )
        self.constraints["01_idx_wmbs_workunit"] = ('CREATE INDEX idx_wmbs_workunit_task'
                                                    ' ON wmbs_workunit(taskid) %s' % tablespaceIndex)
        self.constraints["02_idx_wmbs_workunit"] = ('CREATE INDEX idx_wmbs_workunit_status'
                                                    ' ON wmbs_workunit(status) %s' % tablespaceIndex)

        # Association table between jobs and workunits
        self.create["22wmbs_job_workunit_assoc"] = (
            'CREATE TABLE wmbs_job_workunit_assoc ('
            ' job    INTEGER NOT NULL,'
            ' workunit INTEGER NOT NULL,'
            ' FOREIGN KEY (job)  REFERENCES wmbs_job(id) ON DELETE CASCADE,'
            ' FOREIGN KEY (workunit) REFERENCES wmbs_workunit(id) ON DELETE CASCADE)'
        )
        self.constraints["01_idx_wmbs_job_workunit_assoc"] = ('CREATE INDEX idx_wmbs_job_wu_assoc_job'
                                                              ' ON wmbs_job_workunit_assoc(job) %s' % tablespaceIndex)
        self.constraints["02_idx_wmbs_job_workunit_assoc"] = ('CREATE INDEX idx_wmbs_job_wu_assoc_wu'
                                                              ' ON wmbs_job_workunit_assoc(workunit) %s' % tablespaceIndex)

        # Association table between work units and file/run/lumi triplets
        self.create["23wmbs_frl_workunit_assoc"] = (
            'CREATE TABLE wmbs_frl_workunit_assoc ('
            ' workunit INTEGER NOT NULL,'
            ' firstevent INTEGER DEFAULT 0,'
            ' lastevent INTEGER DEFAULT 0,'
            ' fileid  INTEGER NOT NULL,'
            ' run     INTEGER NOT NULL,'
            ' lumi    INTEGER NOT NULL,'
            ' PRIMARY KEY (workunit, fileid, run, lumi),'
            ' FOREIGN KEY (workunit) REFERENCES wmbs_workunit(id) ON DELETE CASCADE,'
            ' FOREIGN KEY (fileid, run, lumi) REFERENCES wmbs_file_runlumi_map(fileid, run, lumi) ON DELETE CASCADE '
            ')'
        )
        self.constraints["01_idx_wmbs_frl_workunit_assoc"] = ('CREATE INDEX idx_wmbs_frl_wu_assoc_wu'
                                                              ' ON wmbs_frl_workunit_assoc(workunit) %s' % tablespaceIndex)
        self.constraints["02_idx_wmbs_frl_workunit_assoc"] = ('CREATE INDEX idx_wmbs_frl_wu_assoc_frl'
                                                              ' ON wmbs_frl_workunit_assoc(fileid, run, lumi) %s' % tablespaceIndex)

        # Back to other indices and constraints for tables 1-20
        self.constraints["01_idx_wmbs_fileset_files"] = \
            """CREATE INDEX wmbs_fileset_files_idx_fileset ON wmbs_fileset_files(fileset) %s""" % tablespaceIndex

        self.constraints["02_idx_wmbs_fileset_files"] = \
            """CREATE INDEX wmbs_fileset_files_idx_fileid ON wmbs_fileset_files(fileid) %s""" % tablespaceIndex

        self.constraints["01_idx_wmbs_file_runlumi_map"] = \
            """CREATE INDEX wmbs_file_runlumi_map_fileid ON wmbs_file_runlumi_map(fileid) %s""" % tablespaceIndex

        self.constraints["01_idx_wmbs_file_location"] = \
            """CREATE INDEX wmbs_file_location_fileid ON wmbs_file_location(fileid) %s""" % tablespaceIndex

        self.constraints["02_idx_wmbs_file_location"] = \
            """CREATE INDEX wmbs_file_location_pnn ON wmbs_file_location(pnn) %s""" % tablespaceIndex

        self.constraints["01_idx_wmbs_file_parent"] = \
            """CREATE INDEX wmbs_file_parent_parent ON wmbs_file_parent(parent) %s""" % tablespaceIndex

        self.constraints["02_idx_wmbs_file_parent"] = \
            """CREATE INDEX wmbs_file_parent_child ON wmbs_file_parent(child) %s""" % tablespaceIndex

        self.constraints["01_idx_wmbs_workflow_output"] = \
            """CREATE INDEX idx_wmbs_workf_out_workflow ON wmbs_workflow_output(workflow_id) %s""" % tablespaceIndex

        self.constraints["02_idx_wmbs_workflow_output"] = \
            """CREATE INDEX idx_wmbs_workf_out_fileset ON wmbs_workflow_output(output_fileset) %s""" % tablespaceIndex

        self.constraints["03_idx_wmbs_workflow_output"] = \
            """CREATE INDEX idx_wmbs_workf_mout_fileset ON wmbs_workflow_output(merged_output_fileset) %s""" % tablespaceIndex

        self.constraints["01_idx_wmbs_subscription"] = \
            """CREATE INDEX idx_wmbs_subscription_fileset ON wmbs_subscription(fileset) %s""" % tablespaceIndex

        self.constraints["02_idx_wmbs_subscription"] = \
            """CREATE INDEX idx_wmbs_subscription_subtype ON wmbs_subscription(subtype) %s""" % tablespaceIndex

        self.constraints["03_idx_wmbs_subscription"] = \
            """CREATE INDEX idx_wmbs_subscription_workflow ON wmbs_subscription(workflow) %s""" % tablespaceIndex

        self.constraints["01_idx_wmbs_sub_files_acquired"] = \
            """CREATE INDEX idx_wmbs_sub_files_acq_sub ON wmbs_sub_files_acquired(subscription) %s""" % tablespaceIndex

        self.constraints["02_idx_wmbs_sub_files_acquired"] = \
            """CREATE INDEX idx_wmbs_sub_files_acq_file ON wmbs_sub_files_acquired(fileid) %s""" % tablespaceIndex

        self.constraints["01_idx_wmbs_sub_files_available"] = \
            """CREATE INDEX idx_wmbs_sub_files_ava_sub ON wmbs_sub_files_available(subscription) %s""" % tablespaceIndex

        self.constraints["02_idx_wmbs_sub_files_available"] = \
            """CREATE INDEX idx_wmbs_sub_files_ava_file ON wmbs_sub_files_available(fileid) %s""" % tablespaceIndex

        self.constraints["01_idx_wmbs_sub_files_failed"] = \
            """CREATE INDEX idx_wmbs_sub_files_fail_sub ON wmbs_sub_files_failed(subscription) %s""" % tablespaceIndex

        self.constraints["02_idx_wmbs_sub_files_failed"] = \
            """CREATE INDEX idx_wmbs_sub_files_fail_file ON wmbs_sub_files_failed(fileid) %s""" % tablespaceIndex

        self.constraints["01_idx_wmbs_sub_files_complete"] = \
            """CREATE INDEX idx_wmbs_sub_files_comp_sub ON wmbs_sub_files_complete(subscription) %s""" % tablespaceIndex

        self.constraints["02_idx_wmbs_sub_files_complete"] = \
            """CREATE INDEX idx_wmbs_sub_files_comp_file ON wmbs_sub_files_complete(fileid) %s""" % tablespaceIndex

        self.constraints["01_idx_wmbs_sub_jobgroup"] = \
            """CREATE INDEX idx_wmbs_jobgroup_sub ON wmbs_jobgroup(subscription) %s""" % tablespaceIndex

        self.constraints["02_idx_wmbs_sub_jobgroup"] = \
            """CREATE INDEX idx_wmbs_jobgroup_out ON wmbs_jobgroup(output) %s""" % tablespaceIndex

        self.constraints["01_idx_wmbs_job"] = \
            """CREATE INDEX idx_wmbs_job_jobgroup ON wmbs_job(jobgroup) %s""" % tablespaceIndex

        self.constraints["02_idx_wmbs_job"] = \
            """CREATE INDEX idx_wmbs_job_loc ON wmbs_job(location) %s""" % tablespaceIndex

        self.constraints["03_idx_wmbs_job"] = \
            """CREATE INDEX idx_wmbs_job_state ON wmbs_job(state) %s""" % tablespaceIndex

        self.constraints["01_idx_wmbs_job_assoc"] = \
            """CREATE INDEX idx_wmbs_job_assoc_job ON wmbs_job_assoc(job) %s""" % tablespaceIndex

        self.constraints["02_idx_wmbs_job_assoc"] = \
            """CREATE INDEX idx_wmbs_job_assoc_file ON wmbs_job_assoc(fileid) %s""" % tablespaceIndex

        self.constraints["01_idx_wmbs_job_mask"] = \
            """CREATE INDEX idx_wmbs_job_mask_job ON wmbs_job_mask(job) %s""" % tablespaceIndex

        self.constraints["01_idx_wmbs_file_checksums"] = \
            """CREATE INDEX idx_wmbs_file_checksums_type ON wmbs_file_checksums(typeid) %s""" % tablespaceIndex

        self.constraints["01_idx_wmbs_file_checksums"] = \
            """CREATE INDEX idx_wmbs_file_checksums_file ON wmbs_file_checksums(fileid) %s""" % tablespaceIndex

        # The transitions class holds all states and allowed transitions, use
        # that to populate the wmbs_job_state table
        for jobState in Transitions().states():
            jobStateQuery = "INSERT INTO wmbs_job_state (name) VALUES ('%s')" % jobState
            self.inserts["job_state_%s" % jobState] = jobStateQuery

        self.subTypes = [("Processing", 0), ("Merge", 4), ("Harvesting", 5), ("Cleanup", 1),
                         ("LogCollect", 2), ("Skim", 3), ("Production", 0)]
        for pair in self.subTypes:
            subTypeQuery = """INSERT INTO wmbs_sub_types (name, priority)
                                VALUES ('%s', %d)""" % (pair[0], pair[1])
            self.inserts["wmbs_sub_types_%s" % pair[0]] = subTypeQuery

        locationStates = ["Normal", "Down", "Draining", "Aborted"]

        for i in locationStates:
            locationStateQuery = """INSERT INTO wmbs_location_state (name)
                                    VALUES ('%s')""" % i
            self.inserts["wmbs_location_state_%s" % i] = locationStateQuery

        checksumTypes = ['cksum', 'adler32', 'md5']
        for i in checksumTypes:
            checksumTypeQuery = """INSERT INTO wmbs_checksum_type (type) VALUES ('%s') """ % i
            self.inserts["wmbs_checksum_type_%s" % i] = checksumTypeQuery

        return

    def execute(self, conn=None, transaction=None):
        """
        _execute_

        Check to make sure that all required tables have been defined.  If
        everything is in place have the DBCreator make everything.
        """
        for requiredTable in self.requiredTables:
            if requiredTable not in self.create:
                raise WMException("The table '%s' is not defined." % requiredTable, "WMCORE-2")

        DBCreator.execute(self, conn, transaction)
        return True



#!/usr/bin/env python
"""
_DBSBuffer.UpdateAlgo_

Add PSetHash to Algo in DBS Buffer
"""




from WMComponent.DBS3Buffer.MySQL.UpdateAlgo import UpdateAlgo as MySQLUpdateAlgo

class UpdateAlgo(MySQLUpdateAlgo):
    """
    _DBSBuffer.UpdateAlgo_

    Add PSetHash to Algo in DBS Buffer
    """
    pass



#!/usr/bin/env python
"""
_WMBSHelper_

Use WMSpecParser to extract information for creating workflow, fileset, and subscription
"""

from builtins import next, str as newstr, bytes, zip, range
from future.utils import viewitems

import logging
import threading
from collections import defaultdict

from Utils.PythonVersion import PY3
from Utils.Utilities import encodeUnicodeToBytesConditional
from WMComponent.DBS3Buffer.DBSBufferDataset import DBSBufferDataset
from WMComponent.DBS3Buffer.DBSBufferFile import DBSBufferFile
from WMCore.BossAir.BossAirAPI import BossAirAPI, BossAirException
from WMCore.DAOFactory import DAOFactory
from WMCore.DataStructs.File import File as DatastructFile
from WMCore.DataStructs.LumiList import LumiList
from WMCore.DataStructs.Run import Run
from WMCore.JobStateMachine.ChangeState import ChangeState
from WMCore.ResourceControl.ResourceControl import ResourceControl
from WMCore.WMBS.File import File
from WMCore.WMBS.Fileset import Fileset
from WMCore.WMBS.Job import Job
from WMCore.WMBS.Subscription import Subscription
from WMCore.WMBS.Workflow import Workflow
from WMCore.WMConnectionBase import WMConnectionBase
from WMCore.WMException import WMException
from WMCore.WMRuntime.SandboxCreator import SandboxCreator


def wmbsSubscriptionStatus(logger, dbi, conn, transaction):
    """
    Function to return status of wmbs subscriptions
    """
    daoFactory = DAOFactory(package='WMCore.WMBS', logger=logger, dbinterface=dbi)
    action = daoFactory(classname='Monitoring.SubscriptionStatus')
    return action.execute(conn=conn, transaction=transaction)


class WorkQueueWMBSException(WMException):
    """
    Dummy exception class for exceptions raised
    in WMBS Helper.

    TODO: Do something useful

    """

    pass


def killWorkflow(workflowName, jobCouchConfig, bossAirConfig=None):
    """
    _killWorkflow_

    Kill a workflow that is already executing inside the agent.  This will
    mark all incomplete jobs as failed and files that belong to all
    non-cleanup and non-logcollect subscriptions as failed.  The name of the
    JSM couch database and the URL to the database must be passed in as well
    so the state transitions are logged.
    """
    myThread = threading.currentThread()
    daoFactory = DAOFactory(package="WMCore.WMBS",
                            logger=myThread.logger,
                            dbinterface=myThread.dbi)
    killFilesAction = daoFactory(classname="Subscriptions.KillWorkflow")
    killJobsAction = daoFactory(classname="Jobs.KillWorkflow")

    killFilesAction.execute(workflowName=workflowName,
                            conn=myThread.transaction.conn)

    liveJobs = killJobsAction.execute(workflowName=workflowName,
                                      conn=myThread.transaction.conn)

    changeState = ChangeState(jobCouchConfig)

    # Deal with any jobs that are running in the batch system
    # only works if we can start the API
    if bossAirConfig:
        bossAir = BossAirAPI(config=bossAirConfig)
        killableJobs = []
        for liveJob in liveJobs:
            if liveJob["state"].lower() == 'executing':
                # Then we need to kill this on the batch system
                liveWMBSJob = Job(id=liveJob["id"])
                liveWMBSJob.update(liveJob)
                killableJobs.append(liveJob)
        # Now kill them
        try:
            logging.info("Killing %d jobs for workflow: %s", len(killableJobs), workflowName)
            bossAir.kill(jobs=killableJobs, workflowName=workflowName)
        except BossAirException as ex:
            errMsg = "Error while trying to kill running jobs in workflow. Error: %s" % str(ex)
            logging.exception(errMsg)
            # But continue; we need to kill the jobs in the master
            # the batch system will have to take care of itself.

    liveWMBSJobs = defaultdict(list)
    for liveJob in liveJobs:
        if liveJob["state"] == "killed":
            # Then we've killed it already
            continue
        liveWMBSJob = Job(id=liveJob["id"])
        liveWMBSJob.update(liveJob)
        liveWMBSJobs[liveJob["state"]].append(liveWMBSJob)

    for state, jobsByState in viewitems(liveWMBSJobs):
        if len(jobsByState) > 100 and state != "executing":
            # if there are to many jobs skip the couch and dashboard update
            # TODO: couch and dashboard need to be updated or parallel.
            changeState.check("killed", state)
            changeState.persist(jobsByState, "killed", state)
        else:
            changeState.propagate(jobsByState, "killed", state)
    return


def freeSlots(multiplier=1.0, minusRunning=False, allowedStates=None, knownCmsSites=None):
    """
    Get free resources from wmbs.

    Specify multiplier to apply a ratio to the actual numbers.
    minusRunning control if running jobs should be counted
    """
    allowedStates = allowedStates or ['Normal']
    rc_sites = ResourceControl().listThresholdsForCreate()
    thresholds = defaultdict(lambda: 0)
    jobCounts = defaultdict(dict)
    for name, site in viewitems(rc_sites):
        if not site.get('cms_name'):
            logging.warning("Not fetching work for %s, cms_name not defined", name)
            continue
        if knownCmsSites and site['cms_name'] not in knownCmsSites:
            logging.warning(
                "%s doesn't appear to be a known cms site, work may fail to be acquired for it", site['cms_name'])
        if site['state'] not in allowedStates:
            continue
        slots = site['total_slots']
        thresholds[site['cms_name']] += (slots * multiplier)
        if minusRunning:
            jobCounts[site['cms_name']] = dict(
                (k, jobCounts[site['cms_name']].get(k, 0) + site['pending_jobs'].get(k, 0))
                for k in site['pending_jobs'])

    return dict(thresholds), dict(jobCounts)


class WMBSHelper(WMConnectionBase):
    """
    _WMBSHelper_

    Interface between the WorkQueue and WMBS.
    """

    def __init__(self, wmSpec, taskName, blockName=None, mask=None,
                 cachepath='.', commonLocation=None):
        """
        _init_

        Initialize DAOs and other things needed.
        """
        self.block = blockName
        self.mask = mask
        self.wmSpec = wmSpec
        self.topLevelTask = wmSpec.getTask(taskName)
        self.cachepath = cachepath
        self.isDBS = True

        self.topLevelFileset = None
        self.topLevelSubscription = None
        self.topLevelTaskDBSBufferId = None
        self.commonLocation = commonLocation or []

        self.mergeOutputMapping = {}

        # Initiate the pieces you need to run your own DAOs
        WMConnectionBase.__init__(self, "WMCore.WMBS")
        myThread = threading.currentThread()
        self.dbsDaoFactory = DAOFactory(package="WMComponent.DBS3Buffer",
                                        logger=myThread.logger,
                                        dbinterface=myThread.dbi)

        # DAOs from WMBS for file commit
        self.setParentage = self.daofactory(classname="Files.SetParentage")
        self.setFileLocation = self.daofactory(classname="Files.SetLocationForWorkQueue")
        self.setFileAddChecksum = self.daofactory(classname="Files.AddChecksumByLFN")
        self.addFileAction = self.daofactory(classname="Files.Add")
        self.addToFileset = self.daofactory(classname="Files.AddDupsToFileset")
        self.getLocationInfo = self.daofactory(classname="Locations.GetSiteInfo")

        # DAOs from DBSBuffer
        self.dbsCreateFiles = self.dbsDaoFactory(classname="DBSBufferFiles.Add")
        self.dbsSetLocation = self.dbsDaoFactory(classname="DBSBufferFiles.SetLocationByLFN")
        self.dbsInsertLocation = self.dbsDaoFactory(classname="DBSBufferFiles.AddLocation")
        self.dbsSetChecksum = self.dbsDaoFactory(classname="DBSBufferFiles.AddChecksumByLFN")
        self.dbsInsertWorkflow = self.dbsDaoFactory(classname="InsertWorkflow")

        # Added for file creation bookkeeping
        self.dbsFilesToCreate = set()
        self.wmbsFilesToCreate = set()
        self.insertedBogusDataset = -1

        return

    def createSandbox(self):
        """Create the runtime sandbox"""
        sandboxCreator = SandboxCreator()
        sandboxCreator.makeSandbox(self.cachepath, self.wmSpec)

    def createTopLevelFileset(self, topLevelFilesetName=None):
        """
        _createTopLevelFileset_

        Create the top level fileset for the workflow.  If the name of the top
        level fileset is not given create one.
        """
        if topLevelFilesetName is None:
            filesetName = ("%s-%s" % (self.wmSpec.name(),
                                      self.wmSpec.getTopLevelTask()[0].name()))
            if self.block:
                filesetName += "-%s" % self.block
            if self.mask:
                from hashlib import md5
                mask_string = ",".join(["%s=%s" % (x, self.mask[x]) for x in sorted(self.mask)])
                mask_string = encodeUnicodeToBytesConditional(mask_string, condition=PY3)
                filesetName += "-%s" % md5(mask_string).hexdigest()
        else:
            filesetName = topLevelFilesetName

        self.topLevelFileset = Fileset(filesetName)
        self.topLevelFileset.create()
        return

    def outputFilesetName(self, task, outputModuleName, dataTier=''):
        """
        _outputFilesetName_

        Generate an output fileset name for the given task, output module and data tier.
        """
        if task.taskType() == "Merge":
            outputFilesetName = "%s/merged-%s%s" % (task.getPathName(),
                                                    outputModuleName, dataTier)
        else:
            outputFilesetName = "%s/unmerged-%s%s" % (task.getPathName(),
                                                      outputModuleName, dataTier)

        return outputFilesetName

    def createSubscription(self, task, fileset, alternativeFilesetClose=False):
        """
        _createSubscription_

        Create subscriptions in the database.
        This includes workflows in WMBS and DBSBuffer, output maps, datasets
        and phedex subscriptions, and filesets for each task below and including
        the given task.
        """
        self._createSubscriptionsInWMBS(task, fileset, alternativeFilesetClose)

        self._createWorkflowsInDBSBuffer()
        self._createDatasetSubscriptionsInDBSBuffer()

        return

    def _createSubscriptionsInWMBS(self, task, fileset, alternativeFilesetClose=False,
                                   createSandbox=True):
        """
        __createSubscriptionsInWMBS_

        Create subscriptions in WMBS for all the tasks in the spec.  This
        includes filesets, workflows and the output map for each task.
        :param createSandbox: boolean flag to skip (re-)creation of the workload sandbox
        """
        # create runtime sandbox for workflow
        if createSandbox:
            self.createSandbox()

        # FIXME: Let workflow put in values if spec is missing them
        workflow = Workflow(spec=self.wmSpec.specUrl(), owner=self.wmSpec.getOwner()["name"],
                            dn=self.wmSpec.getOwner().get("dn", "unknown"),
                            group=self.wmSpec.getOwner().get("group", "unknown"),
                            owner_vogroup=self.wmSpec.getOwner().get("vogroup", "DEFAULT"),
                            owner_vorole=self.wmSpec.getOwner().get("vorole", "DEFAULT"),
                            name=self.wmSpec.name(), task=task.getPathName(),
                            wfType=self.wmSpec.getDashboardActivity(),
                            alternativeFilesetClose=alternativeFilesetClose,
                            priority=self.wmSpec.priority())
        workflow.create()

        subscription = Subscription(fileset=fileset, workflow=workflow,
                                    split_algo=task.jobSplittingAlgorithm(),
                                    type=task.getPrimarySubType())
        subscription.create()

        ### FIXME: I'm pretty sure we can improve how we handle this site white/black list
        for site in task.siteWhitelist():
            subscription.addWhiteBlackList([{"site_name": site, "valid": True}])

        for site in task.siteBlacklist():
            subscription.addWhiteBlackList([{"site_name": site, "valid": False}])

        if self.topLevelSubscription is None:
            self.topLevelSubscription = subscription
            logging.info("Top level subscription %s created for %s", subscription["id"], self.wmSpec.name())
        else:
            logging.info("Child subscription %s created for %s", subscription["id"], self.wmSpec.name())

        outputModules = task.getOutputModulesForTask()
        ignoredOutputModules = task.getIgnoredOutputModulesForTask()
        for outputModule in outputModules:
            for outputModuleName in sorted(outputModule.listSections_()):
                if outputModuleName in ignoredOutputModules:
                    msg = "%s has %s as IgnoredOutputModule, skipping fileset creation."
                    logging.info(msg, task.getPathName(), outputModuleName)
                    continue
                dataTier = getattr(getattr(outputModule, outputModuleName), "dataTier", '')
                filesetName = self.outputFilesetName(task, outputModuleName, dataTier)
                outputFileset = Fileset(filesetName)
                outputFileset.create()
                outputFileset.markOpen(True)
                mergedOutputFileset = None

                for childTask in task.childTaskIterator():
                    if childTask.data.input.outputModule == outputModuleName:
                        childDatatier = getattr(childTask.data.input, 'dataTier', '')
                        if childTask.taskType() in ["Cleanup", "Merge"] and childDatatier != dataTier:
                            continue
                        elif childTask.taskType() == "Merge" and childDatatier == dataTier:
                            filesetName = self.outputFilesetName(childTask, "Merged", dataTier)
                            mergedOutputFileset = Fileset(filesetName)
                            mergedOutputFileset.create()
                            mergedOutputFileset.markOpen(True)

                            primaryDataset = getattr(getattr(outputModule, outputModuleName), "primaryDataset", None)
                            if primaryDataset is not None:
                                self.mergeOutputMapping[mergedOutputFileset.id] = primaryDataset

                        self._createSubscriptionsInWMBS(childTask, outputFileset,
                                                        alternativeFilesetClose, createSandbox=False)

                if mergedOutputFileset is None:
                    workflow.addOutput(outputModuleName + dataTier, outputFileset,
                                       outputFileset)
                else:
                    workflow.addOutput(outputModuleName + dataTier, outputFileset,
                                       mergedOutputFileset)

        return

    def addMCFakeFile(self):
        """Add a fake file for wmbs to run production over"""
        needed = ['FirstEvent', 'FirstLumi', 'FirstRun', 'LastEvent', 'LastLumi', 'LastRun']
        for key in needed:
            if self.mask and self.mask.get(key) is None:
                msg = 'Invalid value "%s" for %s' % (self.mask.get(key), key)
                raise WorkQueueWMBSException(msg)

        locations = set()
        siteInfo = self.getLocationInfo.execute(conn=self.getDBConn(),
                                                transaction=self.existingTransaction())
        for site in siteInfo:
            if site['pnn'] in self.commonLocation:
                locations.add(site['pnn'])

        if not locations:
            msg = 'No locations to inject Monte Carlo work to, unable to proceed'
            raise WorkQueueWMBSException(msg)
        mcFakeFileName = "MCFakeFile-%s" % self.topLevelFileset.name
        wmbsFile = File(lfn=mcFakeFileName,
                        first_event=self.mask['FirstEvent'],
                        last_event=self.mask['LastEvent'],
                        events=self.mask['LastEvent'] - self.mask['FirstEvent'] + 1,  # inclusive range
                        locations=locations,
                        merged=False,  # merged causes dbs parentage relation
                        )

        if self.mask:
            lumis = list(range(self.mask['FirstLumi'], self.mask['LastLumi'] + 1))  # inclusive range
            wmbsFile.addRun(Run(self.mask['FirstRun'], *lumis))  # assume run number static
        else:
            wmbsFile.addRun(Run(1, 1))

        wmbsFile['inFileset'] = True  # file is not a parent

        logging.debug("WMBS MC Fake File: %s on Location: %s", wmbsFile['lfn'], wmbsFile['newlocations'])

        self.wmbsFilesToCreate.add(wmbsFile)

        totalFiles = self.topLevelFileset.addFilesToWMBSInBulk(self.wmbsFilesToCreate,
                                                               self.wmSpec.name(),
                                                               isDBS=self.isDBS)

        self.topLevelFileset.markOpen(False)
        return totalFiles

    def createSubscriptionAndAddFiles(self, block):
        """
        _createSubscriptionAndAddFiles_

        Create the subscription and add files at one time to
        put everything in one transaction.

        """
        existingTransaction = self.beginTransaction()

        self.createTopLevelFileset()
        try:
            self.createSubscription(self.topLevelTask, self.topLevelFileset)
        except Exception as ex:
            myThread = threading.currentThread()
            myThread.transaction.rollback()
            logging.exception("Failed to create subscription. Error: %s", str(ex))
            raise ex

        if block:
            logging.info('"%s" Injecting block %s (%d files) into wmbs.',
                         self.wmSpec.name(), self.block, len(block['Files']))
            addedFiles = self.addFiles(block)
        else:
            # For MC case
            logging.info('"%s" Injecting production %s:%s:%s - %s:%s:%s (run:lumi:event) into wmbs',
                         self.wmSpec.name(),
                         self.mask['FirstRun'], self.mask['FirstLumi'], self.mask['FirstEvent'],
                         self.mask['LastRun'], self.mask['LastLumi'], self.mask['LastEvent'])
            addedFiles = self.addMCFakeFile()

        self.commitTransaction(existingTransaction)
        logging.info("Transaction committed: %s, for %s", not existingTransaction, self.wmSpec.name())

        # Now that we've created those files, clear the list
        self.dbsFilesToCreate = set()
        self.wmbsFilesToCreate = set()

        return self.topLevelSubscription, addedFiles

    def addFiles(self, block):
        """
        _addFiles_

        create wmbs files from given dbs block.
        as well as run lumi update
        """
        if self.topLevelTask.getInputACDC():
            self.isDBS = False
            logging.info('Adding ACDC files into WMBS for %s', self.wmSpec.name())
            for acdcFile in self.validFiles(block['Files']):
                self._addACDCFileToWMBSFile(acdcFile)
        else:
            self.isDBS = True
            blockPNNs = block['PhEDExNodeNames']
            logging.info('Adding files into WMBS for %s with PNNs: %s', self.wmSpec.name(), blockPNNs)
            for dbsFile in self.validFiles(block['Files']):
                self._addDBSFileToWMBSFile(dbsFile, blockPNNs)

        # Add files to WMBS
        logging.info('Inserting %d files in bulk into WMBS for %s', len(self.wmbsFilesToCreate),
                     self.wmSpec.name())
        totalFiles = self.topLevelFileset.addFilesToWMBSInBulk(self.wmbsFilesToCreate,
                                                               self.wmSpec.name(),
                                                               isDBS=self.isDBS)
        # Add files to DBSBuffer
        logging.info('Inserting %d files in bulk into DBSBuffer for %s', len(self.dbsFilesToCreate),
                     self.wmSpec.name())
        self._createFilesInDBSBuffer()

        # DBS blocks are always closed, so mark fileset as closed as well
        self.topLevelFileset.markOpen(False)
        return totalFiles

    def getMergeOutputMapping(self):
        """
        _getMergeOutputMapping_

        retrieves the relationship between primary
        dataset and merge output fileset ids for
        all merge tasks created
        """
        return self.mergeOutputMapping

    def _createWorkflowsInDBSBuffer(self):
        """
        _createWorkflowsInDBSBuffer_

        Register workflow information and settings in dbsbuffer for all
        tasks that will potentially produce any output in this spec.
        """
        existingTransaction = self.beginTransaction()

        for task in self.wmSpec.listOutputProducingTasks():
            workflow_id = self.dbsInsertWorkflow.execute(self.wmSpec.name(), task,
                                                         self.wmSpec.getBlockCloseMaxWaitTime(),
                                                         self.wmSpec.getBlockCloseMaxFiles(),
                                                         self.wmSpec.getBlockCloseMaxEvents(),
                                                         self.wmSpec.getBlockCloseMaxSize(),
                                                         conn=self.getDBConn(), transaction=self.existingTransaction())
            if task == self.topLevelTask.getPathName():
                self.topLevelTaskDBSBufferId = workflow_id

        self.commitTransaction(existingTransaction)

    def _createDatasetSubscriptionsInDBSBuffer(self):
        """
        _createDatasetSubscriptionsInDBSBuffer_

        Insert the subscriptions defined in the workload for the output
        datasets with the different options.
        """
        existingTransaction = self.beginTransaction()

        subInfo = self.wmSpec.getSubscriptionInformation()
        for dataset in subInfo:
            dbsDataset = DBSBufferDataset(path=dataset)
            dbsDataset.create()
            dbsDataset.addSubscription(subInfo[dataset])
        self.commitTransaction(existingTransaction)
        return

    def _createFilesInDBSBuffer(self):
        """
        _createFilesInDBSBuffer_

        It does the actual job of creating things in DBSBuffer

        """
        if not self.dbsFilesToCreate:
            # Whoops, nothing to do!
            return

        dbsFileTuples = set()
        dbsFileLoc = []
        dbsCksumBinds = []
        locationsToAdd = set()

        # The first thing we need to do is add the datasetAlgo
        # Assume all files in a pass come from one datasetAlgo?
        if self.insertedBogusDataset == -1:
            self.insertedBogusDataset = next(iter(self.dbsFilesToCreate)).insertDatasetAlgo()

        for dbsFile in self.dbsFilesToCreate:
            # Append a tuple in the format specified by DBSBufferFiles.Add
            # Also run insertDatasetAlgo

            lfn = dbsFile['lfn']
            selfChecksums = dbsFile['checksums']

            newTuple = (lfn, dbsFile['size'], dbsFile['events'],
                        self.insertedBogusDataset, dbsFile['status'],
                        self.topLevelTaskDBSBufferId, dbsFile['in_phedex'])

            dbsFileTuples.add(newTuple)

            if len(dbsFile['newlocations']) < 1:
                msg = ''
                msg += "File created without any locations!\n"
                msg += "File lfn: %s\n" % (lfn)
                msg += "Rejecting this group of files in DBS!\n"
                logging.error(msg)
                raise WorkQueueWMBSException(msg)

            for jobLocation in dbsFile['newlocations']:
                locationsToAdd.add(jobLocation)
                dbsFileLoc.append({'lfn': lfn, 'pnn': jobLocation})

            if selfChecksums:
                # If we have checksums we have to create a bind
                # For each different checksum
                for entry in selfChecksums:
                    dbsCksumBinds.append({'lfn': lfn, 'cksum': selfChecksums[entry],
                                          'cktype': entry})

        self.dbsInsertLocation.execute(siteName=locationsToAdd,
                                       conn=self.getDBConn(),
                                       transaction=self.existingTransaction())

        self.dbsCreateFiles.execute(files=dbsFileTuples,
                                    conn=self.getDBConn(),
                                    transaction=self.existingTransaction())

        self.dbsSetLocation.execute(binds=dbsFileLoc,
                                    conn=self.getDBConn(),
                                    transaction=self.existingTransaction())

        if len(dbsCksumBinds) > 0:
            self.dbsSetChecksum.execute(bulkList=dbsCksumBinds,
                                        conn=self.getDBConn(),
                                        transaction=self.existingTransaction())

        return

    def _addToDBSBuffer(self, dbsFile, checksums, locations):
        """
        This step is just for increase the performance for
        Accountant doesn't neccessary to check the parentage
        """
        dbsBuffer = DBSBufferFile(lfn=dbsFile["LogicalFileName"],
                                  size=dbsFile["FileSize"],
                                  events=dbsFile["NumberOfEvents"],
                                  checksums=checksums,
                                  locations=locations,
                                  status="GLOBAL",
                                  inPhedex=1)
        dbsBuffer.setDatasetPath('bogus')
        dbsBuffer.setAlgorithm(appName="cmsRun", appVer="Unknown",
                               appFam="Unknown", psetHash="Unknown",
                               configContent="Unknown")

        self.dbsFilesToCreate.add(dbsBuffer)
        return

    def _addDBSFileToWMBSFile(self, dbsFile, storageElements, inFileset=True):
        """
        There are two assumptions made to make this method behave properly,
        1. DBS returns only one level of ParentList.
           If DBS returns multiple level of parentage, it will be still get handled.
           However that might not be what we wanted. In that case, restrict to one level.
        2. Assumes parents files are in the same location as child files.
           This is not True in general case, but workquue should only select work only
           where child and parent files are in the same location
        """
        # TODO get dbsFile with lumi event information
        wmbsParents = []
        dbsFile.setdefault("ParentList", [])
        for parent in dbsFile["ParentList"]:
            wmbsParents.append(self._addDBSFileToWMBSFile(parent,
                                                          storageElements, inFileset=False))

        checksums = {}
        if dbsFile.get('Checksum'):
            checksums['cksum'] = dbsFile['Checksum']
        if dbsFile.get('Adler32'):
            checksums['adler32'] = dbsFile['Adler32']

        wmbsFile = File(lfn=dbsFile["LogicalFileName"],
                        size=dbsFile["FileSize"],
                        events=dbsFile["NumberOfEvents"],
                        checksums=checksums,
                        # TODO: need to get list of parent lfn
                        parents=wmbsParents,
                        locations=set(storageElements))

        for lumi in dbsFile['LumiList']:
            if isinstance(lumi['LumiSectionNumber'], list):
                lumiSecList = (list(zip(lumi['LumiSectionNumber'], lumi['EventCount']))
                               if 'EventCount' in lumi else lumi['LumiSectionNumber'])
                run = Run(lumi['RunNumber'], lumiSecList)
            else:
                lumiSecTuple = ((lumi['LumiSectionNumber'], lumi['EventCount'])
                                if 'EventCount' in lumi else lumi['LumiSectionNumber'])
                run = Run(lumi['RunNumber'], lumiSecTuple)
            wmbsFile.addRun(run)

        self._addToDBSBuffer(dbsFile, checksums, storageElements)

        logging.debug("WMBS File: %s on Location: %s", wmbsFile['lfn'], wmbsFile['newlocations'])

        wmbsFile['inFileset'] = bool(inFileset)
        self.wmbsFilesToCreate.add(wmbsFile)

        return wmbsFile

    def _convertACDCFileToDBSFile(self, acdcFile):
        """
        convert ACDCFiles to dbs file format
        """
        dbsFile = {}
        dbsFile["LogicalFileName"] = acdcFile["lfn"]
        dbsFile["FileSize"] = acdcFile["size"]
        dbsFile["NumberOfEvents"] = acdcFile["events"]
        return dbsFile

    def _addACDCFileToWMBSFile(self, acdcFile, inFileset=True):
        """
        adds the ACDC files into WMBS database
        """
        wmbsParents = []
        # If file is merged, then it will be the parent of whatever output that
        # process this job (it twists my mind!). Meaning, block below can be skipped
        if int(acdcFile.get("merged", 0)) == 0 and acdcFile["parents"]:
            # set the parentage for all the unmerged parents
            for parent in acdcFile["parents"]:
                if parent.startswith("/store/unmerged/") or parent.startswith("MCFakeFile"):
                    logging.warning("WMBS ACDC skipped parent invalid file: %s", parent)
                    continue
                logging.debug("WMBS ACDC Parent File: %s", parent)
                parent = self._addACDCFileToWMBSFile(DatastructFile(lfn=parent,
                                                                    locations=acdcFile["locations"],
                                                                    merged=True),
                                                     inFileset=False)
                wmbsParents.append(parent)

        # pass empty check sum since it won't be updated to dbs anyway
        checksums = {}
        wmbsFile = File(lfn=str(acdcFile["lfn"]),
                        size=acdcFile["size"],
                        events=acdcFile["events"],
                        first_event=acdcFile.get('first_event', 0),
                        last_event=acdcFile.get('last_event', 0),
                        checksums=checksums,
                        parents=wmbsParents,
                        locations=acdcFile["locations"],
                        merged=acdcFile.get('merged', True))

        ## TODO need to get the lumi lists
        for run in acdcFile['runs']:
            wmbsFile.addRun(run)

        if not acdcFile["lfn"].startswith("/store/unmerged") or wmbsParents:
            # only add to DBSBuffer if is not unmerged file or it has parents.
            dbsFile = self._convertACDCFileToDBSFile(acdcFile)
            self._addToDBSBuffer(dbsFile, checksums, acdcFile["locations"])

        logging.debug("WMBS ACDC File: %s on Location: %s", wmbsFile['lfn'], wmbsFile['newlocations'])

        wmbsFile['inFileset'] = bool(inFileset)

        self.wmbsFilesToCreate.add(wmbsFile)

        return wmbsFile

    def validFiles(self, files):
        """
        Apply lumi mask and or run white/black list and return files which have
        one or more of the requested lumis
        """
        runWhiteList = self.topLevelTask.inputRunWhitelist()
        runBlackList = self.topLevelTask.inputRunBlacklist()
        lumiMask = self.topLevelTask.getLumiMask()

        blackMask = None
        if lumiMask:  # We have a lumiMask, so use it and modify with run white/black list
            if runWhiteList:
                lumiMask.selectRuns(runWhiteList)
            if runBlackList:
                lumiMask.removeRuns(runBlackList)
        elif runWhiteList:  # We have a run whitelist, subtract off blacklist
            lumiMask = LumiList(runs=runWhiteList)
            if runBlackList:  # We only have a blacklist, so make a black mask out of it instead
                lumiMask.removeRuns(runBlackList)
        else:
            lumiMask = None
            if runBlackList:
                blackMask = LumiList(runs=runBlackList)

        results = []
        for f in files:
            if isinstance(f, (newstr, bytes)) or "LumiList" not in f:
                results.append(f)
                continue

            # Create a LumiList from the WMBS info
            runLumis = {}
            for x in f['LumiList']:
                if x['RunNumber'] in runLumis:
                    runLumis[x['RunNumber']].extend(x['LumiSectionNumber'])
                else:
                    runLumis[x['RunNumber']] = x['LumiSectionNumber']
            fileLumiList = LumiList(runsAndLumis=runLumis)

            if lumiMask:
                if fileLumiList & lumiMask:  # At least one lumi from file is in lumiMask
                    results.append(f)
            elif blackMask:
                if fileLumiList - blackMask:  # At least one lumi from file is not in blackMask
                    results.append(f)
            else:  # There is effectively no mask
                results.append(f)

        return results



#!/usr/bin/env python
"""
_LoadFromFilesetWorkflow_

MySQL implementation of Subscription.LoadFromFilesetWorkflow
"""

__all__ = []



from WMCore.Database.DBFormatter import DBFormatter

class LoadFromFilesetWorkflow(DBFormatter):
    sql = """SELECT wmbs_subscription.id, fileset, workflow, split_algo,
                    wmbs_sub_types.name, last_update FROM wmbs_subscription
               INNER JOIN wmbs_sub_types ON
                 wmbs_subscription.subtype = wmbs_sub_types.id
             WHERE fileset = :fileset AND workflow = :workflow"""

    def formatDict(self, result):
        """
        _formatDict_

        Cast the id, fileset, workflow and last_update columns to integers
        since formatDict() turns everything into strings.
        """
        formattedResult = DBFormatter.formatDict(self, result)[0]
        formattedResult["id"] = int(formattedResult["id"])
        formattedResult["fileset"] = int(formattedResult["fileset"])
        formattedResult["workflow"] = int(formattedResult["workflow"])
        formattedResult["last_update"] = int(formattedResult["last_update"])
        formattedResult["type"] = formattedResult["name"]
        del formattedResult["name"]
        return formattedResult

    def execute(self, fileset = None, workflow = None, conn = None,
                transaction = False):
        result = self.dbi.processData(self.sql, {"fileset": fileset,
                                                 "workflow": workflow},
                                      conn = conn, transaction = transaction)
        return self.formatDict(result)



#!/usr/bin/env python
"""
_Jobs_

MySQL implementation of Subscriptions.Jobs
"""

__all__ = []



from WMCore.Database.DBFormatter import DBFormatter

class Jobs(DBFormatter):
    sql = """SELECT id FROM wmbs_job WHERE jobgroup IN
             (SELECT id FROM wmbs_jobgroup WHERE subscription = :subscription)"""

    def formatDict(self, results):
        """
        _formatDict_

        Cast the id column to an integer since formatDict() turns everything
        into strings.
        """
        formattedResults = DBFormatter.formatDict(self, results)

        for formattedResult in formattedResults:
            formattedResult["id"] = int(formattedResult["id"])

        return formattedResults

    def execute(self, subscription = 0, conn = None, transaction = False):
        results = self.dbi.processData(self.sql, {"subscription": subscription},
                                       conn = conn, transaction = transaction)
        return self.formatDict(results)



#!/usr/bin/env python

"""
MySQL implementation of AddChecksumByLFN
"""





from WMCore.Database.DBFormatter import DBFormatter

class AddChecksumByLFN(DBFormatter):
    sql = """INSERT IGNORE INTO wmbs_file_checksums (fileid, typeid, cksum)
             SELECT (SELECT id FROM wmbs_file_details WHERE lfn = :lfn),
              (SELECT id FROM wmbs_checksum_type WHERE type = :cktype),
              :cksum FROM dual"""

    def execute(self, lfn = None, cktype = None, cksum = None, bulkList = None, conn = None,
                transaction = False):

        if bulkList:
            binds = bulkList
        else:
            binds = {'lfn': lfn, 'cktype': cktype, 'cksum': cksum}

        result = self.dbi.processData(self.sql, binds,
                         conn = conn, transaction = transaction)

        return



"""
_GetDeletableBlocks_

Oracle implementation of PhEDExInjector.GetDeletableBlocks

"""

from __future__ import division
from __future__ import print_function

from WMComponent.RucioInjector.Database.MySQL.GetDeletableBlocks import GetDeletableBlocks as MySQLBase


class GetDeletableBlocks(MySQLBase):
    pass



#!/usr/bin/env python
"""
_GetMigratedBlocks_

"""

from __future__ import division

from WMCore.Database.DBFormatter import DBFormatter


class GetMigratedBlocks(DBFormatter):
    sql = """SELECT dbsbuffer_block.blockname AS blockname,
                    dbsbuffer_dataset.path AS dataset,
                    dbsbuffer_location.pnn AS location
                    FROM dbsbuffer_block
               INNER JOIN
                 (SELECT block_id, MIN(dbsbuffer_file.id) file_id FROM dbsbuffer_file
                  GROUP BY block_id) file_block_assoc ON
                 dbsbuffer_block.id = file_block_assoc.block_id
               INNER JOIN dbsbuffer_file ON
                 file_block_assoc.file_id = dbsbuffer_file.id
               INNER JOIN dbsbuffer_file_location ON
                 dbsbuffer_file.id = dbsbuffer_file_location.filename
               INNER JOIN dbsbuffer_location ON
                 dbsbuffer_file_location.location = dbsbuffer_location.id
               INNER JOIN dbsbuffer_algo_dataset_assoc ON
                 dbsbuffer_file.dataset_algo = dbsbuffer_algo_dataset_assoc.id
               INNER JOIN dbsbuffer_dataset ON
                 dbsbuffer_algo_dataset_assoc.dataset_id = dbsbuffer_dataset.id
             WHERE dbsbuffer_block.status = 'InDBS'
             AND NOT EXISTS (SELECT dbf.id FROM dbsbuffer_file dbf
                              WHERE dbf.block_id = dbsbuffer_block.id
                              AND dbf.in_phedex = 0)"""

    def formatDict(self, result):
        """
        _formatDict_

        Format the query results into something that resembles the XML format
        PhEDEx expects for injection:

        {"location1":
          {"dataset1":
            {"block1": {"is-open": "n", "files": []}}}}
        """
        dictResult = DBFormatter.formatDict(self, result)

        formattedResult = {}
        for row in dictResult:
            if row["location"] not in formattedResult:
                formattedResult[row["location"]] = {}

            locationDict = formattedResult[row["location"]]
            if row["dataset"] not in locationDict:
                locationDict[row["dataset"]] = {}

            datasetDict = locationDict[row["dataset"]]
            if row["blockname"] not in datasetDict:
                datasetDict[row["blockname"]] = {"is-open": "n",
                                                 "files": []}

        return formattedResult

    def execute(self, conn=None, transaction=False):
        result = self.dbi.processData(self.sql, conn=conn,
                                      transaction=transaction)
        return self.formatDict(result)



#!/usr/bin/env python
"""
_GetLocation_

Oracle implementation of Jobs.Location
"""




from WMCore.WMBS.MySQL.Jobs.GetLocation import GetLocation as MySQLGetLocation

class GetLocation(MySQLGetLocation):
    """
    _GetLocation_

    Retrieve all files that are associated with the given job from the
    database.
    """



"""
_Create_

Base class for creating the BossAir database
"""
from __future__ import print_function

import threading

from WMCore.Database.DBCreator import DBCreator
from WMCore.WMException import WMException


class Create(DBCreator):
    """
    This should create the BossAir schema; since they don't do it.

    """

    def __init__(self, logger=None, dbi=None, params=None):
        """
        _init_

        Call the DBCreator constructor and create the list of required tables.
        """
        myThread = threading.currentThread()

        if logger is None:
            logger = myThread.logger
        if dbi is None:
            dbi = myThread.dbi

        tablespaceIndex = ""
        if params:
            if "tablespace_index" in params:
                tablespaceIndex = "USING INDEX TABLESPACE %s" % params["tablespace_index"]

        DBCreator.__init__(self, logger, dbi)

        self.requiredTables = ["01bl_status", "02bl_runjob"]

        self.create['01bl_status'] = \
            """CREATE TABLE bl_status
                (
                id            INT  auto_increment,
                name          VARCHAR(255),
                PRIMARY KEY (id),
                UNIQUE (name)
                )
                ENGINE=InnoDB ROW_FORMAT=DYNAMIC;
            """

        self.create['02bl_runjob'] = \
            """CREATE TABLE bl_runjob
               (
               id            INT   auto_increment,
               wmbs_id       INT,
               grid_id       VARCHAR(255),
               bulk_id       VARCHAR(255),
               status        CHAR(1)   DEFAULT '1',
               sched_status  INT NOT NULL,
               retry_count   INT,
               status_time   INT,
               location      INT,
               user_id       INT,
               PRIMARY KEY (id),
               FOREIGN KEY (wmbs_id) REFERENCES wmbs_job(id) ON DELETE CASCADE,
               FOREIGN KEY (sched_status) REFERENCES bl_status(id),
               FOREIGN KEY (user_id) REFERENCES wmbs_users(id) ON DELETE CASCADE,
               FOREIGN KEY (location) REFERENCES wmbs_location(id) ON DELETE CASCADE,
               UNIQUE (retry_count, wmbs_id)
               )
               ENGINE=InnoDB ROW_FORMAT=DYNAMIC;

            """

        self.constraints["01_idx_bl_runjob"] = \
            """CREATE INDEX idx_bl_runjob_wmbs ON bl_runjob(wmbs_id) %s""" % tablespaceIndex

        self.constraints["02_idx_bl_runjob"] = \
            """CREATE INDEX idx_bl_runjob_status ON bl_runjob(sched_status) %s""" % tablespaceIndex

        self.constraints["03_idx_bl_runjob"] = \
            """CREATE INDEX idx_bl_runjob_users ON bl_runjob(user_id) %s""" % tablespaceIndex

        self.constraints["04_idx_bl_runjob"] = \
            """CREATE INDEX idx_bl_runjob_location ON bl_runjob(location) %s""" % tablespaceIndex

        return

    def execute(self, conn=None, transaction=None):
        """
        _execute_

        Check to make sure that all required tables have been defined.  If
        everything is in place have the DBCreator make everything.
        """
        for requiredTable in self.requiredTables:
            if requiredTable not in self.create:
                raise WMException("The table '%s' is not defined." % \
                                  requiredTable, "WMCORE-2")

        try:
            DBCreator.execute(self, conn, transaction)
            return True
        except Exception as e:
            print("ERROR: %s" % e)
            return False



#!/usr/bin/env python
"""
_PluginInterface_

Base class for AnalyticsDataCollector plug-ins

"""

from builtins import object
class PluginInterface(object):
    """Interface for policies"""
    def __init__(self, **args):
        """ initialize args if needed"""
        pass

    def __call__(self, requestDocs, localSummaryCouchDB, centralRequestCouchDB):
        """
        this needs to be overwritten
        requestDocs is request info collected from other resources, which has couchdb doc format.
        localSummaryCouchDB, centralWMStatsCouchDB are WMCore.Services.WMStats.WMStatsWriter instances
        for local wmagent_summary db and central wmstats couchdb respectfully.
        If data needed to be written to wmstats, write to local it will be replicated,
        Only in case data doesn't exist in local, write to central wmstats couchdb.
        requestDocs can be modified and it will be updated outside the plug in,
        So don't push requestDocs to couchdb directly here.
        """
        msg = "%s.__call__ is not implemented" % self.__class__.__name__
        raise NotImplementedError(msg)



#!/usr/bin/env python
"""
_DeleteCheck_

Oracle implementation of DeleteCheckFiles
"""

from WMCore.WMBS.MySQL.Files.DeleteCheck import DeleteCheck as MySQLDeleteCheck

class DeleteCheck(MySQLDeleteCheck):
    pass



#!/usr/bin/env python
"""

Oracle implementation of GetByID
"""




from WMComponent.DBS3Buffer.MySQL.DBSBufferFiles.GetByID import GetByID as MySQLGetByID

class GetByID(MySQLGetByID):
    """

    Oracle implementation of GetByID
    """
    pass



#!/usr/bin/env python
"""
_Step.Executor.LogCollect_

Implementation of an Executor for a  LogCollect step.
"""
from __future__ import print_function

import datetime
import logging
import os
import signal
import socket
import tarfile

from Utils.IteratorTools import grouper
from WMCore.Algorithms.Alarm import Alarm, alarmHandler
from WMCore.Storage.DeleteMgr import DeleteMgr
from WMCore.Storage.StageInMgr import StageInMgr
from WMCore.Storage.StageOutError import StageOutFailure
from WMCore.Storage.StageOutMgr import StageOutMgr
from WMCore.WMRuntime.Tools.Scram import Scram, getSingleScramArch, isCMSSWSupported
from WMCore.WMSpec.Steps.Executor import Executor
from WMCore.WMSpec.Steps.WMExecutionFailure import WMExecutionFailure
from WMCore.Storage.SiteLocalConfig import loadSiteLocalConfig


class LogCollect(Executor):
    """
    _LogCollect_

    Execute a LogCollect Step

    """

    def pre(self, emulator=None):
        """
        _pre_

        Pre execution checks

        """
        # Are we using an emulator?
        if emulator is not None:
            return emulator.emulatePre(self.step)

        logging.info("Steps.Executors.%s.pre called", self.__class__.__name__)
        return None

    def execute(self, emulator=None):
        """
        _execute_

        """
        # Are we using emulators again?
        if emulator is not None:
            return emulator.emulate(self.step, self.job)

        logging.info("Steps.Executors.%s.execute called", self.__class__.__name__)

        scramCommand = self.step.application.setup.scramCommand
        cmsswVersion = self.step.application.setup.cmsswVersion
        scramArch = getSingleScramArch(self.step.application.setup.scramArch)
        overrideCatalog = getattr(self.step.application, 'overrideCatalog', None)

        overrides = {}
        if hasattr(self.step, 'override'):
            overrides = self.step.override.dictionary_()
            if overrides.get('logRedirectSiteLocalConfig', None):
                siteCfg = loadSiteLocalConfig()
                overrides.update(siteCfg.localStageOut)

        # Set wait to over an hour
        waitTime = overrides.get('waitTime', 3600 + (self.step.retryDelay * self.step.retryCount))

        # hardcode CERN EOS T2_CH_CERN stageout parameters
        eosStageOutParams = {}
        eosStageOutParams['command'] = overrides.get('command', "xrdcp")
        eosStageOutParams['option'] = overrides.get('option', "--wma-disablewriterecovery")
        eosStageOutParams['phedex-node'] = overrides.get('phedex-node', "T2_CH_CERN")
        eosStageOutParams['lfn-prefix'] = overrides.get('lfn-prefix', "root://eoscms.cern.ch//eos/cms")

        try:
            eosStageOutMgr = StageOutMgr(**eosStageOutParams)
            stageInMgr = StageInMgr()
            deleteMgr = DeleteMgr()
        except Exception as ex:
            msg = "Unable to load StageOut/Delete Impl: %s" % str(ex)
            logging.error(msg)
            raise WMExecutionFailure(60312, "MgrImplementationError", msg)

        # prepare output tar file
        taskName = self.report.getTaskName().split('/')[-1]
        host = socket.gethostname().split('.')[0]
        tarName = '%s-%s-%s-%i-logs.tar' % (self.report.data.workload, taskName, host, self.job["counter"])
        tarLocation = os.path.join(self.stepSpace.location, tarName)

        # Supported by any release beyond CMSSW_8_X, however DaviX is broken until CMSSW_10_4_X
        # see: https://github.com/cms-sw/cmssw/issues/25292
        useEdmCopyUtil = True
        if isCMSSWSupported(cmsswVersion, "CMSSW_10_4_0"):
            pass
        elif scramArch.startswith('slc7_amd64_'):
            msg = "CMSSW too old or not fully functional to support edmCopyUtil, using CMSSW_10_4_0 instead"
            logging.warning(msg)
            cmsswVersion = "CMSSW_10_4_0"
            scramArch = "slc7_amd64_gcc820"
        elif scramArch.startswith('slc6_amd64_'):
            msg = "CMSSW too old or not fully functional to support edmCopyUtil, using CMSSW_10_4_0 instead"
            logging.warning(msg)
            cmsswVersion = "CMSSW_10_4_0"
            scramArch = "slc6_amd64_gcc700"
        else:
            useEdmCopyUtil = False
        logging.info("Using edmCopyUtil: %s", useEdmCopyUtil)

        # setup Scram needed to run edmCopyUtil
        if useEdmCopyUtil:
            scram = Scram(
                    command=scramCommand,
                    version=cmsswVersion,
                    initialise=self.step.application.setup.softwareEnvironment,
                    directory=self.step.builder.workingDir,
                    architecture=scramArch,
            )
            logging.info("Running scram")
            try:
                projectOutcome = scram.project()
            except Exception as ex:
                msg = "Exception raised while running scram.\n"
                msg += str(ex)
                logging.critical("Error running SCRAM")
                logging.critical(msg)
                raise WMExecutionFailure(50513, "ScramSetupFailure", msg)

            if projectOutcome > 0:
                msg = scram.diagnostic()
                logging.critical("Error running SCRAM")
                logging.critical(msg)
                raise WMExecutionFailure(50513, "ScramSetupFailure", msg)
            runtimeOutcome = scram.runtime()
            if runtimeOutcome > 0:
                msg = scram.diagnostic()
                logging.critical("Error running SCRAM")
                logging.critical(msg)
                raise WMExecutionFailure(50513, "ScramSetupFailure", msg)

        # iterate through input files
        localLogs = []
        deleteLogArchives = []
        if useEdmCopyUtil:
            numberOfFilesPerCopy = 10
        else:
            numberOfFilesPerCopy = 1
        for logs in grouper(self.job["input_files"], numberOfFilesPerCopy):

            copyCommand = "env X509_USER_PROXY=%s edmCopyUtil" % os.environ.get('X509_USER_PROXY', None)

            # specify TFC if necessary
            if overrideCatalog:
                copyCommand += " -c %s" % overrideCatalog

            for log in logs:
                copyCommand += " %s" % log['lfn']
            copyCommand += " %s" % self.step.builder.workingDir

            # give up after timeout of 1 minute per input file
            signal.signal(signal.SIGALRM, alarmHandler)
            signal.alarm(60 * numberOfFilesPerCopy)

            filesCopied = False
            try:
                if useEdmCopyUtil:
                    logging.info("Running edmCopyUtil")
                    retval = scram(copyCommand)
                    if retval == 0:
                        filesCopied = True
                else:
                    logging.info("Running stageIn")
                    for log in logs:
                        fileInfo = {"LFN": log['lfn']}
                        logArchive = stageInMgr(**fileInfo)
                        if logArchive:
                            filesCopied = True
            except Alarm:
                logging.error("Indefinite hang during edmCopyUtil/stageIn of logArchives")
            except StageOutFailure:
                logging.error("Unable to stageIn logArchives")
            except Exception:
                raise

            signal.alarm(0)

            if filesCopied:
                for log in logs:
                    localLogs.append(os.path.join(self.step.builder.workingDir, os.path.basename(log['lfn'])))
                    deleteLogArchives.append(log)
                    self.report.addInputFile(sourceName="logArchives", lfn=log['lfn'])
            else:
                logging.error("Unable to copy logArchives to local disk")
                for log in logs:
                    self.report.addSkippedFile(log['lfn'], None)

        # create tarfile if any logArchive copied in
        if localLogs:
            with tarfile.open(tarLocation, 'w:') as tarFile:
                for log in localLogs:
                    path = log.split('/')
                    tarFile.add(name=log,
                                arcname=os.path.join(path[-3], path[-2], path[-1]))
                    os.remove(log)
        else:
            msg = "Unable to copy any logArchives to local disk"
            logging.error(msg)
            raise WMExecutionFailure(60312, "LogCollectError", msg)

        # now staging out the LogCollect tarfile
        logging.info("Staging out LogCollect tarfile to EOS (skipping CASTOR)")
        now = datetime.datetime.now()
        lfn = "/store/logs/prod/%i/%.2i/%s/%s/%s" % (now.year, now.month, "WMAgent",
                                                     self.report.data.workload,
                                                     os.path.basename(tarLocation))

        tarInfo = {'LFN': lfn,
                   'PFN': tarLocation,
                   'PNN': None,
                   'GUID': None}

        # perform mandatory stage out to CERN EOS
        signal.signal(signal.SIGALRM, alarmHandler)
        signal.alarm(waitTime)
        try:
            eosStageOutMgr(tarInfo)
        except Alarm:
            msg = "Indefinite hang during stageOut of LogCollect to EOS"
            logging.error(msg)
            raise WMExecutionFailure(60409, "LogCollectTimeout", msg)
        except Exception as ex:
            msg = "Unable to stageOut LogCollect to Castor:\n"
            msg += str(ex)
            logging.error(msg)
            raise WMExecutionFailure(60408, "LogCollectStageOutError", msg)
        signal.alarm(0)

        # add to job report
        self.report.addOutputFile(outputModule="LogCollect", aFile=tarInfo)
        outputRef = getattr(self.report.data, self.stepName)
        outputRef.output.pfn = tarInfo['PFN']
        outputRef.output.location = tarInfo['PNN']
        outputRef.output.lfn = tarInfo['LFN']

        # we got this far, delete ALL input files assigned to this job
        for log in self.job["input_files"]:
            # give up after timeout of 1 minutes
            signal.signal(signal.SIGALRM, alarmHandler)
            signal.alarm(60)
            try:
                fileToDelete = {'LFN': log['lfn'],
                                'PFN': None,
                                'PNN': None,
                                'StageOutCommand': None}
                deleteMgr(fileToDelete=fileToDelete)
            except Alarm:
                logging.error("Indefinite hang during delete of logArchive")
            except Exception as ex:
                logging.error("Unable to delete logArchive: %s", ex)
            signal.alarm(0)

        return

    def post(self, emulator=None):
        """
        _post_

        Post execution checkpointing

        """
        # Another emulator check
        if emulator is not None:
            return emulator.emulatePost(self.step)

        logging.info("Steps.Executors.%s.post called", self.__class__.__name__)
        return None



#/usr/bin/env python2.4
"""
_Destroy_

MySQL implementation of BossAir.Destroy
"""

__revision__ = "$Id: Destroy.py,v 1.2 2010/05/10 13:34:33 spigafi Exp $"
__version__ = "$Revision: 1.2 $"

import threading

from WMCore.Database.DBCreator import DBCreator

from WMCore.WMException import WMException
from WMCore.WMExceptions import WMEXCEPTION

from WMCore.BossAir.Oracle.Create import Create

class Destroy(DBCreator):
    """
    BossAir.Destroy
    """

    def __init__(self, logger = None, dbi = None):
        """
        _init_

        Call the DBCreator constructor and create the list of required tables.
        """
        myThread = threading.currentThread()

        if logger == None:
            logger = myThread.logger
        if dbi == None:
            dbi = myThread.dbi

        DBCreator.__init__(self, logger, dbi)

        self.delete['01bl_runjob']    = "DROP TABLE bl_runjob"
        self.delete['02bl_status']    = "DROP TABLE bl_status"


        j = 50
        for i in Create.sequence_tables:
            seqname = '%s_SEQ' % i
            self.create["%s%s" % (j, seqname)] = \
                           "DROP SEQUENCE %s"  % seqname

        return



#!/usr/bin/env python
"""
_GetLocationForJobs_

MySQL implementation of JobGroup.GetLocationsForJobs
"""

__all__ = []

import logging

from WMCore.Database.DBFormatter import DBFormatter

from future.utils import listvalues

class GetLocationsForJobs(DBFormatter):
    """
    _GetLocationsForJobs_

    Working under the current jobSplitting assumption, namely that each jobGroup contains
    only jobs that can run at the same SET of sites, we can find the jobs the jobGroup can
    run at by selecting them from the files.
    """
    sql = """SELECT DISTINCT site_name FROM wmbs_location wl
          INNER JOIN wmbs_location_pnns wlpnn ON wlpnn.location = wl.id
          INNER JOIN wmbs_file_location wfl ON wfl.pnn = wlpnn.pnn
          INNER JOIN wmbs_job_assoc wja ON wja.fileid = wfl.fileid
          INNER JOIN wmbs_job wj ON wj.id = wja.job
          WHERE wj.jobgroup = :jobgroup"""

    def format(self, result):

        newResult = []
        # First, actually get something useful
        modifiedResult = result[0].fetchall()

        for res in modifiedResult:
            # Should only have one entry (one site_name per site)
            tmp = listvalues(res)[0]
            if not tmp in newResult:
                newResult.append(tmp)

        return newResult

    def execute(self, id, conn=None, transaction=False):
        """
        _execute_

        Execute the SQL for the given job ID and then format and return
        the result.
        """

        if id == -1:
            logging.error("JobGroup.GetLocationsForJobs got unspecified jobGroup ID")
            return []

        result = self.dbi.processData(self.sql, {"jobgroup": id}, conn=conn,
                                      transaction=transaction)

        return self.format(result)



# -*- coding: utf-8 -*-
"""
_GetJobSlots_

Oracle implementation of Locations.GetJobSlots
Created on Mon Jun 18 15:28:46 2012

@author: dballest
"""

from WMCore.WMBS.MySQL.Locations.GetJobSlots import GetJobSlots as MySQLGetJobSlots

class GetJobSlots(MySQLGetJobSlots):
    """
    _GetJobSlots_

    Same as MySQL version
    """
    pass
